# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
import os
import tempfile
from multiprocessing import Pool

import numpy as np
from huggingface_hub.errors import HFValidationError
from transformers.utils import cached_file

try:
    from staticvectors import Database, StaticVectors  # type:ignore

    STATICVECTORS = True
except ImportError:
    STATICVECTORS = False

from ...pipelines import Tokenizer
from ..base import Vectors

logger = logging.getLogger(__name__)

# pylint: disable=W0603
PARAMETERS, VECTORS = None, None


def create(config, scoring):
    """
    Multiprocessing helper method. Creates a global embeddings object to be accessed in a new subprocess.

    Args:
        config: vector configuration
        scoring: scoring instance
    """

    global PARAMETERS
    global VECTORS

    PARAMETERS, VECTORS = (config, scoring, None), None


def transform(document):
    """
    Multiprocessing helper method. Transforms document into an embeddings vector.

    Args:
        document: (id, data, tags)

    Returns:
        (id, embedding)
    """

    global VECTORS
    if not VECTORS:
        VECTORS = WordVectors(*PARAMETERS)

    return (document[0], VECTORS.transform(document))


class WordVectors(Vectors):
    """
    Builds vectors using weighted word embeddings.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a WordVectors model.

        Args:
            path: input path

        Returns:
            True if this is a WordVectors model, False otherwise
        """

        if WordVectors.isdatabase(path):
            return True

        try:
            path = cached_file(path_or_repo_id=path, filename="config.json")
            if path:
                with open(path, encoding="utf-8") as f:
                    config = json.load(f)
                    return config.get("model_type") == "staticvectors"

        except (HFValidationError, OSError):
            pass

        return False

    @staticmethod
    def isdatabase(path):
        """
        Checks if this is a SQLite database file which is the file format used for word vectors databases.

        Args:
            path: path to check

        Returns:
            True if this is a SQLite database
        """

        return isinstance(path, str) and STATICVECTORS and Database.isdatabase(path)

    def __init__(self, config, scoring, models):
        if not STATICVECTORS:
            raise ImportError('staticvectors is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        return StaticVectors(path)

    def encode(self, data, category=None):
        embeddings = []
        for tokens in data:
            if isinstance(tokens, str):
                tokenlist = Tokenizer.tokenize(tokens)
                tokens = tokenlist if tokenlist else [tokens]

            weights = self.scoring.weights(tokens) if self.scoring else None

            # pylint: disable=E1133
            if weights and [x for x in weights if x > 0]:
                embedding = np.average(self.lookup(tokens), weights=np.array(weights, dtype=np.float32), axis=0)
            else:
                embedding = np.mean(self.lookup(tokens), axis=0)

            embeddings.append(embedding)

        return np.array(embeddings, dtype=np.float32)

    def index(self, documents, batchsize=500, checkpoint=None):
        parallel = self.config.get("parallel", True)
        parallel = os.cpu_count() if parallel and isinstance(parallel, bool) else int(parallel)

        if not parallel:
            return super().index(documents, batchsize)

        ids, dimensions, batches, stream = [], None, 0, None

        args = (self.config, self.scoring)

        with Pool(parallel, initializer=create, initargs=args) as pool:
            with tempfile.NamedTemporaryFile(mode="wb", suffix=".npy", delete=False) as output:
                stream = output.name
                embeddings = []
                for uid, embedding in pool.imap(transform, documents, self.encodebatch):
                    if not dimensions:
                        dimensions = embedding.shape[0]

                    ids.append(uid)
                    embeddings.append(embedding)

                    if len(embeddings) == batchsize:
                        np.save(output, np.array(embeddings, dtype=np.float32), allow_pickle=False)
                        batches += 1

                        embeddings = []

                if embeddings:
                    np.save(output, np.array(embeddings, dtype=np.float32), allow_pickle=False)
                    batches += 1

        return (ids, dimensions, batches, stream)

    def lookup(self, tokens):
        """
        Queries word vectors for given list of input tokens.

        Args:
            tokens: list of tokens to query

        Returns:
            word vectors array
        """

        return self.model.embeddings(tokens)

    def tokens(self, data):
        return data



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect

from transformers import pipeline

from ..modeling import Models
from ..utilities import Resolver
from .tensors import Tensors


class HFPipeline(Tensors):
    """
    Light wrapper around Hugging Face Transformers pipeline component for selected tasks. Adds support for model
    quantization and minor interface changes.
    """

    def __init__(self, task, path=None, quantize=False, gpu=False, model=None, **kwargs):
        """
        Loads a new pipeline model.

        Args:
            task: pipeline task or category
            path: optional path to model, accepts Hugging Face model hub id, local path or (model, tokenizer) tuple.
                  uses default model for task if not provided
            quantize: if model should be quantized, defaults to False
            gpu: True/False if GPU should be enabled, also supports a GPU device id
            model: optional existing pipeline model to wrap
            kwargs: additional keyword arguments to pass to pipeline model
        """

        if model:
            self.pipeline = model.pipeline if isinstance(model, HFPipeline) else model
        else:
            deviceid = Models.deviceid(gpu) if "device_map" not in kwargs else None
            device = Models.device(deviceid) if deviceid is not None else None

            modelargs, kwargs = self.parseargs(**kwargs)

            if isinstance(path, list | tuple):
                config = path[1] if path[1] and isinstance(path[1], str) else None

                model = Models.load(path[0], config, task)

                self.pipeline = pipeline(
                    task, model=model, tokenizer=path[1], device=device, model_kwargs=modelargs, **kwargs
                )
            else:
                self.pipeline = pipeline(task, model=path, device=device, model_kwargs=modelargs, **kwargs)

            if deviceid == -1 and quantize:
                # pylint: disable=E1101
                self.pipeline.model = self.quantize(self.pipeline.model)

        Models.checklength(self.pipeline.model, self.pipeline.tokenizer)

    def parseargs(self, **kwargs):
        """
        Inspects the pipeline method and splits kwargs into model args and pipeline args.

        Args:
            kwargs: all keyword arguments

        Returns:
            (model args, pipeline args)
        """

        args = inspect.getfullargspec(pipeline).args

        dtype = kwargs.get("torch_dtype")
        if dtype and isinstance(dtype, str) and dtype != "auto":
            kwargs["torch_dtype"] = Resolver()(dtype)

        return (
            {arg: value for arg, value in kwargs.items() if arg not in args},
            {arg: value for arg, value in kwargs.items() if arg in args},
        )

    def maxlength(self):
        """
        Gets the max length to use for generate calls.

        Returns:
            max length
        """

        return Models.maxlength(self.pipeline.model, self.pipeline.tokenizer)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import msgpack
from msgpack import Unpacker
from msgpack.exceptions import ExtraData

from .base import Serialize
from .errors import SerializeError


class MessagePack(Serialize):
    """
    MessagePack serialization.
    """

    def __init__(self, streaming=False, **kwargs):
        super().__init__()

        self.streaming = streaming

        self.kwargs = kwargs

    def loadstream(self, stream):
        try:
            return Unpacker(stream, **self.kwargs) if self.streaming else msgpack.unpack(stream)
        except ExtraData as e:
            raise SerializeError(e) from e

    def savestream(self, data, stream):
        msgpack.pack(data, stream)

    def loadbytes(self, data):
        return msgpack.unpackb(data)

    def savebytes(self, data):
        return msgpack.packb(data)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .action import Action
from .autoid import AutoId
from .configuration import Configuration
from .documents import Documents
from .functions import Functions
from .indexes import Indexes
from .indexids import IndexIds
from .reducer import Reducer
from .stream import Stream
from .transform import Transform

__all__ = (
    "Action",
    "AutoId",
    "Configuration",
    "Documents",
    "Functions",
    "IndexIds",
    "Indexes",
    "Reducer",
    "Stream",
    "Transform",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers import AutoModel, AutoModelForQuestionAnswering, AutoModelForSequenceClassification
from transformers.models.auto.tokenization_auto import TOKENIZER_MAPPING


class Registry:
    """
    Methods to register models and fully support pipelines.
    """

    @staticmethod
    def register(model, config=None):
        """
        Registers a model with auto model and tokenizer configuration to fully support pipelines.

        Args:
            model: model to register
            config: config class name
        """

        config = config if config else model.__class__

        if hasattr(model.__class__, "config_class") and not model.__class__.config_class:
            model.__class__.config_class = config

        for mapping in [AutoModel, AutoModelForQuestionAnswering, AutoModelForSequenceClassification]:
            mapping.register(config, model.__class__)

        if hasattr(model, "config") and type(model.config) not in TOKENIZER_MAPPING:
            TOKENIZER_MAPPING.register(type(model.config), type(model.config).__name__)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO

try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from .base import Encoder


class ImageEncoder(Encoder):
    """
    Encodes and decodes Image objects as compressed binary content, using the original image's algorithm.
    """

    def __init__(self):
        """
        Creates a new ImageEncoder.
        """

        if not PIL:
            raise ImportError('ImageEncoder is not available - install "database" extra to enable')

    def encode(self, obj):
        output = BytesIO()

        obj.save(output, format=obj.format, quality="keep")

        return output.getvalue()

    def decode(self, data):
        return Image.open(BytesIO(data)) if data else None



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
vector_embeddings package provides classes and utilities for managing, indexing, and searching embeddings.

This package exposes the main Embeddings interface and related tools for embedding index management, search, and transformation.

Modules exported:
    - Embeddings: Main class for embedding management and search.
    - Action, AutoId, Configuration, Documents, Functions, Indexes, IndexIds, Reducer, Stream, Transform: Indexing and transformation utilities.
    - Explain, Ids, IndexNotFoundError, Query, Scan, Search, Terms: Search and query utilities.
"""

from .base import Embeddings
from .embedding_index import (
    Action,
    AutoId,
    Configuration,
    Documents,
    Functions,
    Indexes,
    IndexIds,
    Reducer,
    Stream,
    Transform,
)
from .embedding_search import Explain, Ids, IndexNotFoundError, Query, Scan, Search, Terms

__all__ = (
    "Action",
    "AutoId",
    "Configuration",
    "Documents",
    "Embeddings",
    "Explain",
    "Functions",
    "Ids",
    "IndexIds",
    "IndexNotFoundError",
    "Indexes",
    "Query",
    "Reducer",
    "Scan",
    "Search",
    "Stream",
    "Terms",
    "Transform",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import VectoreIndex
from .dense_index import HNSW, ANNFactory, Annoy, Faiss, NumPy, PGVector, Torch
from .sparse_index import IVFSparse, PGSparse, SparseANNFactory

__all__ = (
    "HNSW",
    "ANNFactory",
    "Annoy",
    "Faiss",
    "IVFSparse",
    "NumPy",
    "PGSparse",
    "PGVector",
    "SparseANNFactory",
    "Torch",
    "VectoreIndex",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...serialization import SerializeFactory
from .base import Encoder


class SerializeEncoder(Encoder):
    """
    Encodes and decodes objects using the internal serialize package.
    """

    def __init__(self, method):
        super().__init__()

        self.serializer = SerializeFactory.create(method)

    def encode(self, obj):
        return self.serializer.savebytes(obj)

    def decode(self, data):
        return self.serializer.loadbytes(data)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Scan class for scanning indexes for query matches and the Clause class for parsing query clause parameters.
"""


class Scan:
    """
    Scans indexes for query matches and executes scans for lists of queries.

    Args:
        search (Callable): Index search function.
        limit (int): Maximum results.
        weights (Any): Default hybrid score weights.
        index (str): Default index name.
    """

    def __init__(self, search, limit, weights, index):
        """
        Creates a new scan instance.

        Args:
            search: index search function
            limit: maximum results
            weights: default hybrid score weights
            index: default index name
        """

        self.search = search

        self.limit = limit

        self.candidates = None

        self.weights = weights

        self.index = index

    def __call__(self, queries, parameters):
        """
        Executes a scan for a list of queries.

        Args:
            queries: list of queries to run
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            list of (id, score) per query
        """

        results = {}

        default = None

        for index, iqueries in self.parse(queries, parameters).items():
            candidates = [query.candidates for query in iqueries if query.candidates]
            if not candidates and not default:
                default = self.default(queries)

            candidates = max(candidates) if candidates else default

            weights = [query.weights for query in iqueries if query.weights is not None]
            weights = max(weights) if weights else self.weights

            index = index if index else self.index

            for x, result in enumerate(self.search([query.text for query in iqueries], candidates, weights, index)):
                results[iqueries[x].uid] = (iqueries[x].qid, result)

        return [result for _, result in sorted(results.items())]

    def parse(self, queries, parameters):
        """
        Parse index query clauses from a list of parsed queries.

        Args:
            queries: list of parsed queries
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            index query clauses grouped by index
        """

        results, uid = {}, 0
        for x, query in enumerate(queries):
            if "similar" in query:
                for params in query["similar"]:
                    if parameters and parameters[x]:
                        params = self.bind(params, parameters[x])

                    clause = Clause(uid, x, params)

                    if clause.index not in results:
                        results[clause.index] = []

                    results[clause.index].append(clause)
                    uid += 1

        return results

    def bind(self, similar, parameters):
        """
        Resolves bind parameters for a similar function call.

        Args:
            similar: similar function call arguments
            parameters: bind parameters

        Returns:
            similar function call arguments with resolved bind parameters
        """

        resolved = []
        for p in similar:
            if isinstance(p, str) and p.startswith(":") and p[1:] in parameters:
                resolved.append(parameters[p[1:]])
            else:
                resolved.append(p)

        return resolved

    def default(self, queries):
        """
        Derives the default number of candidates. The number of candidates are the number of results to bring back
        from index queries. This is an optional argument to similar() clauses.

        For a single query filter clause, the default is the query limit. With multiple filtering clauses, the default is
        10x the query limit. This ensures that limit results are still returned with additional filtering after an index.

        Args:
            queries: list of queries

        Returns:
            default candidate list size
        """

        multitoken = any(query.get("where") and len(query["where"].split()) > 1 for query in queries)
        return self.limit * 10 if multitoken else self.limit


class Clause:
    """
    Parses and stores query clause parameters for index scans.

    Args:
        uid (int): Query clause id.
        qid (int): Query id clause is a part of.
        params (Any): Query parameters to parse.
    """

    def __init__(self, uid, qid, params):
        """
        Creates a new query clause.

        Args:
            uid: query clause id
            qid: query id clause is a part of
            params: query parameters to parse
        """

        self.uid, self.qid = uid, qid
        self.text, self.index = params[0], None
        self.candidates, self.weights = None, None

        if len(params) > 1:
            self.parse(params[1:])

    def parse(self, params):
        """
        Parses clause parameters into this instance.

        Args:
            params: query clause parameters
        """

        for param in params:
            if (isinstance(param, str) and param.isdigit()) or isinstance(param, int):
                self.candidates = int(param)

            elif (isinstance(param, str) and param.replace(".", "").isdigit()) or isinstance(param, float):
                self.weights = float(param)

            else:
                self.index = param



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Template module
"""

from string import Formatter

from ...utilities import TemplateFormatter
from .file import Task


class TemplateTask(Task):
    """
    Task that generates text from a template and task inputs. Templates can be used to prepare data for a number of tasks
    including generating large language model (LLM) prompts.
    """

    def register(self, template=None, rules=None, strict=True):
        """
        Read template parameters.

        Args:
            template: prompt template
            rules: parameter rules
            strict: requires all task inputs to be consumed by template, defaults to True
        """

        # pylint: disable=W0201
        self.template = template if template else self.defaulttemplate()

        self.rules = rules if rules else self.defaultrules()

        self.formatter = TemplateFormatter() if strict else Formatter()

    def prepare(self, element):
        match = self.match(element)
        if match:
            return match

        if self.template:
            if isinstance(element, dict):
                return self.formatter.format(self.template, **element)

            if isinstance(element, tuple):
                return self.formatter.format(self.template, **{f"arg{i}": x for i, x in enumerate(element)})

            return self.formatter.format(self.template, text=element)

        return element

    def defaulttemplate(self):
        """
        Generates a default template for this task. Base method returns None.

        Returns:
            default template
        """

        return None

    def defaultrules(self):
        """
        Generates a default rules for this task. Base method returns an empty dictionary.

        Returns:
            default rules
        """

        return {}

    def match(self, element):
        """
        Check if element matches any processing rules.

        Args:
            element: input element

        Returns:
            matching value if found, None otherwise
        """

        if self.rules and isinstance(element, dict):
            for key, value in self.rules.items():
                if element[key] == value:
                    return element[key]

        return None


class RagTask(TemplateTask):
    """
    Template task that prepares input for a rag pipeline.
    """

    def prepare(self, element):
        if isinstance(element, dict):
            params = dict(element)
            params.pop("query", None)
            params["text"] = params.pop("question")

            element["question"] = super().prepare(params)
            return element

        return {"query": element, "question": super().prepare(element)}



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch

from .numpy import NumPy


class Torch(NumPy):
    """
    Builds an VectoreIndex index backed by a PyTorch array.
    """

    def __init__(self, config):
        super().__init__(config)

        self.all, self.cat, self.dot, self.zeros = torch.all, torch.cat, torch.mm, torch.zeros
        self.argsort, self.xor, self.clip = torch.argsort, torch.bitwise_xor, torch.clip

    def tensor(self, array):
        if isinstance(array, np.ndarray):
            array = torch.from_numpy(array)

        return array.cuda() if torch.cuda.is_available() else array

    def numpy(self, array):
        return array.cpu().numpy()

    def totype(self, array, dtype):
        return array.long() if dtype == np.int64 else array

    def settings(self):
        return {"torch": torch.__version__}



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..modeling import Models
from .tensors import Tensors


class HFModel(Tensors):
    """
    Pipeline backed by a Hugging Face Transformers model.
    """

    def __init__(self, path=None, quantize=False, gpu=False, batch=64):
        """
        Creates a new HFModel.

        Args:
            path: optional path to model, accepts Hugging Face model hub id or local path,
                  uses default model for task if not provided
            quantize: if model should be quantized, defaults to False
            gpu: True/False if GPU should be enabled, also supports a GPU device id
            batch: batch size used to incrementally process content
        """

        self.path = path

        self.quantization = quantize

        self.deviceid = Models.deviceid(gpu)
        self.device = Models.device(self.deviceid)

        self.batchsize = batch

    def prepare(self, model):
        """
        Prepares a model for processing. Applies dynamic quantization if necessary.

        Args:
            model: input model

        Returns:
            model
        """

        if self.deviceid == -1 and self.quantization:
            model = self.quantize(model)

        return model

    def tokenize(self, tokenizer, texts):
        """
        Tokenizes text using tokenizer. This method handles overflowing tokens and automatically splits
        them into separate elements. Indices of each element is returned to allow reconstructing the
        transformed elements after running through the model.

        Args:
            tokenizer: Tokenizer
            texts: list of text

        Returns:
            (tokenization result, indices)
        """

        batch, positions = [], []
        for x, text in enumerate(texts):
            elements = [t + " " for t in text.split("\n") if t]
            batch.extend(elements)
            positions.extend([x] * len(elements))

        tokens = tokenizer(batch, padding=True)

        inputids, attention, indices = [], [], []
        for x, ids in enumerate(tokens["input_ids"]):
            if len(ids) > tokenizer.model_max_length:
                ids = [i for i in ids if i != tokenizer.pad_token_id]

                for chunk in self.batch(ids, tokenizer.model_max_length - 1):
                    if chunk[-1] != tokenizer.eos_token_id:
                        chunk.append(tokenizer.eos_token_id)

                    mask = [1] * len(chunk)

                    if len(chunk) < tokenizer.model_max_length:
                        pad = tokenizer.model_max_length - len(chunk)
                        chunk.extend([tokenizer.pad_token_id] * pad)
                        mask.extend([0] * pad)

                    inputids.append(chunk)
                    attention.append(mask)
                    indices.append(positions[x])
            else:
                inputids.append(ids)
                attention.append(tokens["attention_mask"][x])
                indices.append(positions[x])

        tokens = {"input_ids": inputids, "attention_mask": attention}

        # pylint: disable=E1102
        return ({name: self.tensor(tensor).to(self.device) for name, tensor in tokens.items()}, indices)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re
import types

import numpy as np
import torch

logger = logging.getLogger(__name__)


class Task:
    """
    Base class for all workflow tasks.
    """

    def __init__(
        self,
        action=None,
        select=None,
        unpack=True,
        column=None,
        merge="hstack",
        initialize=None,
        finalize=None,
        concurrency=None,
        onetomany=True,
        **kwargs,
    ):
        """
        Creates a new task. A task defines two methods, type of data it accepts and the action to execute
        for each data element. Action is a callable function or list of callable functions.

        Args:
            action: action(s) to execute on each data element
            select: filter(s) used to select data to process
            unpack: if data elements should be unpacked or unwrapped from (id, data, tag) tuples
            column: column index to select if element is a tuple, defaults to all
            merge: merge mode for joining multi-action outputs, defaults to hstack
            initialize: action to execute before processing
            finalize: action to execute after processing
            concurrency: sets concurrency method when execute instance available
                         valid values: "thread" for thread-based concurrency, "process" for process-based concurrency
            onetomany: if one-to-many data transformations should be enabled, defaults to True
            kwargs: additional keyword arguments
        """

        if not action:
            action = []
        elif not isinstance(action, list):
            action = [action]

        self.action = action
        self.select = select
        self.unpack = unpack
        self.column = column
        self.merge = merge
        self.initialize = initialize
        self.finalize = finalize
        self.concurrency = concurrency
        self.onetomany = onetomany

        if hasattr(self, "register"):
            self.register(**kwargs)
        elif kwargs:
            kwargs = ", ".join(f"'{kw}'" for kw in kwargs)
            raise TypeError(f"__init__() got unexpected keyword arguments: {kwargs}")

    def __call__(self, elements, executor=None):
        """
        Executes action for a list of data elements.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        if isinstance(elements, list):
            return self.filteredrun(elements, executor)

        return self.run(elements, executor)

    def filteredrun(self, elements, executor):
        """
        Executes a filtered run, which will tag all inputs with a process id, filter elements down to elements the
        task can handle and execute on that subset. Items not selected for processing will be returned unmodified.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        indexed = list(enumerate(elements))

        data = [(x, self.upack(element)) for x, element in indexed if self.accept(self.upack(element, True))]

        ids = [x for x, _ in data]

        results = self.execute([self.prepare(element) for _, element in data], executor)

        if self.merge:
            elements = self.filteredpack(results, indexed, ids)
        else:
            elements = [self.filteredpack(r, indexed, ids) for r in results]

        return elements

    def filteredpack(self, results, indexed, ids):
        """
        Processes and packs results back into original input elements.

        Args:
            results: task results
            indexed: original elements indexed by process id
            ids: process ids accepted by this task

        Returns:
            packed elements
        """

        elements = []
        for x, element in indexed:
            if x in ids:
                result = results[ids.index(x)]

                if isinstance(result, OneToMany):
                    elements.extend([self.pack(element, r) for r in result])
                else:
                    elements.append(self.pack(element, result))
            else:
                elements.append(element)

        return elements

    def run(self, elements, executor):
        """
        Executes a task run for elements. A standard run processes all elements.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        results = self.execute(elements, executor)

        if isinstance(results, list):
            elements = []
            for result in results:
                if isinstance(result, OneToMany):
                    elements.extend(result)
                else:
                    elements.append(result)

            return elements

        return results

    def accept(self, element):
        """
        Determines if this task can handle the input data format.

        Args:
            element: input data element

        Returns:
            True if this task can process this data element, False otherwise
        """

        return (
            (isinstance(element, str) and re.search(self.select, element.lower()))
            if element is not None and self.select
            else True
        )

    def upack(self, element, force=False):
        """
        Unpacks data for processing.

        Args:
            element: input data element
            force: if True, data is unpacked even if task has unpack set to False

        Returns:
            data
        """

        if (self.unpack or force) and isinstance(element, tuple) and len(element) > 1:
            return element[1]

        return element

    def pack(self, element, data):
        """
        Packs data after processing.

        Args:
            element: transformed data element
            data: item to pack element into

        Returns:
            packed data
        """

        if self.unpack and isinstance(element, tuple) and len(element) > 1:
            if isinstance(data, tuple) and (len(self.action) <= 1 or self.merge != "hstack"):
                return data

            element = list(element)
            element[1] = data
            return tuple(element)

        return data

    def prepare(self, element):
        """
        Method that allows downstream tasks to prepare data element for processing.

        Args:
            element: input data element

        Returns:
            data element ready for processing
        """

        return element

    def execute(self, elements, executor):
        """
        Executes action(s) on elements.

        Args:
            elements: list of data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        if self.action:
            outputs = []
            for x, action in enumerate(self.action):
                index = self.column[x] if isinstance(self.column, dict) else self.column
                inputs = [self.extract(e, index) for e in elements] if index is not None else elements

                outputs.append((action, inputs) if executor else self.process(action, inputs))

            if executor:
                outputs = executor.run(self.concurrency, self.process, outputs)

            return self.postprocess(outputs)

        return elements

    def extract(self, element, index):
        """
        Extracts a column from element by index if the element is a tuple.

        Args:
            element: input element
            index: column index

        Returns:
            extracted column
        """

        if isinstance(element, tuple):
            if not self.unpack and len(element) == 3 and isinstance(element[1], tuple):
                return (element[0], element[1][index], element[2])

            return element[index]

        return element

    def process(self, action, inputs):
        """
        Executes action using inputs as arguments.

        Args:
            action: callable object
            inputs: action inputs

        Returns:
            action outputs
        """

        logger.debug("Inputs: %s", inputs)

        outputs = action(inputs)

        if isinstance(outputs, types.GeneratorType):
            outputs = list(outputs)

        logger.debug("Outputs: %s", outputs)

        return outputs

    def postprocess(self, outputs):
        """
        Runs post process routines after a task action.

        Args:
            outputs: task outputs

        Returns:
            postprocessed outputs
        """

        if len(self.action) == 1:
            return self.single(outputs[0])

        if not self.merge:
            return outputs

        if self.merge == "vstack":
            return self.vstack(outputs)
        if self.merge == "concat":
            return self.concat(outputs)

        return self.hstack(outputs)

    def single(self, outputs):
        """
        Post processes and returns single action outputs.

        Args:
            outputs: outputs from a single task

        Returns:
            post processed outputs
        """

        if self.onetomany and isinstance(outputs, list):
            outputs = [OneToMany(output) if isinstance(output, list) else output for output in outputs]

        return outputs

    def vstack(self, outputs):
        """
        Merges outputs row-wise. Returns a list of lists which will be interpreted as a one to many transformation.

        Row-wise merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Row Merge => [[a1, a2], [b1, b2], [c1, c2]] = [a1, a2, b1, b2, c1, c2]

        Args:
            outputs: task outputs

        Returns:
            list of aggregated/zipped outputs as one to many transforms (row-wise)
        """

        if all(isinstance(output, np.ndarray) for output in outputs):
            return np.concatenate(np.stack(outputs, axis=1))

        # pylint: disable=E1101
        if all(torch.is_tensor(output) for output in outputs):
            return torch.cat(tuple(torch.stack(outputs, axis=1)))

        merge = []
        for x in zip(*outputs, strict=False):
            combine = []
            for y in x:
                if isinstance(y, list):
                    combine.extend(y)
                else:
                    combine.append(y)

            merge.append(OneToMany(combine))

        return merge

    def hstack(self, outputs):
        """
        Merges outputs column-wise. Returns a list of tuples which will be interpreted as a one to one transformation.

        Column-wise merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Column Merge => [(a1, a2), (b1, b2), (c1, c2)]

        Args:
            outputs: task outputs

        Returns:
            list of aggregated/zipped outputs as tuples (column-wise)
        """

        if all(isinstance(output, np.ndarray) for output in outputs):
            return np.stack(outputs, axis=1)

        # pylint: disable=E1101
        if all(torch.is_tensor(output) for output in outputs):
            return torch.stack(outputs, axis=1)

        return list(zip(*outputs, strict=False))

    def concat(self, outputs):
        """
        Merges outputs column-wise and concats values together into a string. Returns a list of strings.

        Concat merge example (2 actions)

          Inputs: [a, b, c]

          Outputs => [[a1, b1, c1], [a2, b2, c2]]

          Concat Merge => [(a1, a2), (b1, b2), (c1, c2)] => ["a1. a2", "b1. b2", "c1. c2"]

        Args:
            outputs: task outputs

        Returns:
            list of concat outputs
        """

        return [". ".join([str(y) for y in x if y]) for x in self.hstack(outputs)]


class OneToMany:
    """
    Encapsulates list output for a one to many transformation.
    """

    def __init__(self, values):
        """
        Creates a new OneToMany transformation.

        Args:
            values: list of outputs
        """

        self.values = values

    def __iter__(self):
        return self.values.__iter__()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ...serialization import SerializeFactory
from ..base import VectoreIndex


class NumPy(VectoreIndex):
    """
    Builds an VectoreIndex index backed by a NumPy array.
    """

    def __init__(self, config):
        super().__init__(config)

        self.all, self.cat, self.dot, self.zeros = np.all, np.concatenate, np.dot, np.zeros
        self.argsort, self.xor, self.clip = np.argsort, np.bitwise_xor, np.clip

        quantize = self.config.get("quantize")
        self.qbits = quantize if quantize and isinstance(quantize, int) and not isinstance(quantize, bool) else None

    def load(self, path):
        try:
            self.backend = self.tensor(np.load(path, allow_pickle=False))
        except ValueError:
            self.backend = self.tensor(SerializeFactory.create("pickle").load(path))

    def index(self, embeddings):
        self.backend = self.tensor(embeddings)

        self.config["offset"] = embeddings.shape[0]
        self.metadata(self.settings())

    def append(self, embeddings):
        self.backend = self.cat((self.backend, self.tensor(embeddings)), axis=0)

        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        ids = [x for x in ids if x < self.backend.shape[0]]

        self.backend[ids] = self.tensor(self.zeros((len(ids), self.backend.shape[1])))

    def search(self, queries, limit):
        if self.qbits:
            scores = self.hammingscore(queries)
        else:
            scores = self.dot(self.tensor(queries), self.backend.T)

        ids = self.argsort(-scores)[:, :limit]

        results = []
        for x, score in enumerate(scores):
            results.append(list(zip(ids[x].tolist(), score[ids[x]].tolist(), strict=False)))

        return results

    def count(self):
        return self.backend[~self.all(self.backend == 0, axis=1)].shape[0]

    def save(self, path):
        with open(path, "wb") as handle:
            np.save(handle, self.numpy(self.backend), allow_pickle=False)

    def tensor(self, array):
        """
        Handles backend-specific code such as loading to a GPU device.

        Args:
            array: data array

        Returns:
            array with backend-specific logic applied
        """

        return array

    def numpy(self, array):
        """
        Handles backend-specific code to convert an array to numpy

        Args:
            array: data array

        Returns:
            numpy array
        """

        return array

    def totype(self, array, dtype):
        """
        Casts array to dtype.

        Args:
            array: input array
            dtype: dtype

        Returns:
            array cast as dtype
        """

        return np.int64(array) if dtype == np.int64 else array

    def settings(self):
        """
        Returns settings for this array.

        Returns:
            dict
        """

        return {"numpy": np.__version__}

    def hammingscore(self, queries):
        """
        Calculates a hamming distance score.

        This is defined as:

            score = 1.0 - (hamming distance / total number of bits)

        Args:
            queries: queries array

        Returns:
            scores
        """

        table = 1 << np.arange(8)
        table = self.tensor(np.array([np.count_nonzero(x & table) for x in np.arange(256)]))

        delta = self.xor(self.tensor(queries[:, None]), self.backend)

        delta = self.totype(delta, np.int64)

        return self.clip(1.0 - (table[delta].sum(axis=2) / (self.config["dimensions"] * 8)), 0.0, 1.0)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

from ...serialization import SerializeFactory


class Documents:
    """
    Streams documents to temporary storage. Allows queuing large volumes of content for later indexing.
    """

    def __init__(self):
        """
        Creates a new documents stream.
        """

        self.documents = None
        self.batch = 0
        self.size = 0

        self.serializer = SerializeFactory.create("pickle", allowpickle=True)

    def __len__(self):
        """
        Returns total number of queued documents.
        """

        return self.size

    def __iter__(self):
        """
        Streams all queued documents.
        """

        self.documents.close()

        with open(self.documents.name, "rb") as queue:
            for _ in range(self.batch):
                documents = self.serializer.loadstream(queue)

                yield from documents

    def add(self, documents):
        """
        Adds a batch of documents for indexing.

        Args:
            documents: list of (id, data, tag) tuples

        Returns:
            documents
        """

        # pylint: disable=R1732
        if not self.documents:
            self.documents = tempfile.NamedTemporaryFile(mode="wb", suffix=".docs", delete=False)

        self.serializer.savestream(documents, self.documents)
        self.batch += 1
        self.size += len(documents)

        return documents

    def close(self):
        """
        Closes and resets this instance. New sets of documents can be added with additional calls to add.
        """

        os.remove(self.documents.name)

        self.documents = None
        self.batch = 0
        self.size = 0



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .annoy import Annoy
from .factory import ANNFactory
from .faiss import Faiss
from .hnsw import HNSW
from .numpy import NumPy
from .pgvector import PGVector
from .torch import Torch

__all__ = ("HNSW", "ANNFactory", "Annoy", "Faiss", "NumPy", "PGVector", "Torch")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Task
from .console import ConsoleTask
from .export import ExportTask
from .factory import TaskFactory
from .file import FileTask
from .image import ImageTask
from .retrieve import RetrieveTask
from .service import ServiceTask
from .storage import StorageTask
from .stream import StreamTask
from .template import Formatter, RagTask, TemplateFormatter, TemplateTask
from .template import RagTask as ExtractorTask
from .url import UrlTask
from .workflow import WorkflowTask

__all__ = (
    "ConsoleTask",
    "ExportTask",
    "ExtractorTask",
    "FileTask",
    "Formatter",
    "ImageTask",
    "RagTask",
    "RetrieveTask",
    "ServiceTask",
    "StorageTask",
    "StreamTask",
    "Task",
    "TaskFactory",
    "TemplateFormatter",
    "TemplateTask",
    "UrlTask",
    "WorkflowTask",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from annoy import AnnoyIndex

    ANNOY = True
except ImportError:
    ANNOY = False

from ..base import VectoreIndex


# pylint: disable=W0223
class Annoy(VectoreIndex):
    """
    Builds an VectoreIndex index using the Annoy library.
    """

    def __init__(self, config):
        super().__init__(config)

        if not ANNOY:
            raise ImportError('Annoy is not available - install "ann" extra to enable')

    def load(self, path):
        self.backend = AnnoyIndex(self.config["dimensions"], self.config["metric"])
        self.backend.load(path)

    def index(self, embeddings):
        self.config["metric"] = "dot"

        self.backend = AnnoyIndex(self.config["dimensions"], self.config["metric"])

        for x in range(embeddings.shape[0]):
            self.backend.add_item(x, embeddings[x])

        ntrees = self.setting("ntrees", 10)
        self.backend.build(ntrees)

        self.metadata({"ntrees": ntrees})

    def search(self, queries, limit):
        searchk = self.setting("searchk", -1)

        results = []
        for query in queries:
            ids, scores = self.backend.get_nns_by_vector(query, n=limit, search_k=searchk, include_distances=True)

            results.append(list(zip(ids, scores, strict=False)))

        return results

    def count(self):
        return self.backend.get_n_items()

    def save(self, path):
        self.backend.save(path)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .action import Action


class Transform:
    """
    Executes a transform. Processes a stream of documents, loads batches into enabled data stores
    and vectorizes documents.
    """

    def __init__(self, embeddings, action, checkpoint=None):
        """
        Creates a new transform.

        Args:
            embeddings: embeddings instance
            action: index action
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        self.embeddings = embeddings
        self.action = action
        self.checkpoint = checkpoint

        self.config = embeddings.config
        self.delete = embeddings.delete
        self.model = embeddings.model
        self.database = embeddings.database
        self.graph = embeddings.graph
        self.indexes = embeddings.indexes
        self.scoring = embeddings.scoring if embeddings.issparse() else None

        self.offset = embeddings.config.get("offset", 0) if action == Action.UPSERT else 0
        self.batch = embeddings.config.get("batch", 1024)

        quantize = embeddings.config.get("quantize")
        self.qbits = quantize if isinstance(quantize, int) and not isinstance(quantize, bool) else None

        columns = embeddings.config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        self.indexing = embeddings.model or embeddings.scoring

        self.deletes = set()

    def __call__(self, documents, buffer):
        """
        Processes an iterable collection of documents, handles any iterable including generators.

        This method loads a stream of documents into enabled data stores and vectorizes
        documents into an embeddings array.

        Args:
            documents: iterable of (id, data, tags)
            buffer: file path used for memmap buffer

        Returns:
            (document ids, dimensions, embeddings)
        """

        ids, dimensions, embeddings = None, None, None

        if self.model:
            ids, dimensions, embeddings = self.vectors(documents, buffer)
        else:
            ids = self.ids(documents)

        return (ids, dimensions, embeddings)

    def vectors(self, documents, buffer):
        """
        Runs a vectors transform operation when dense indexing is enabled.

        Args:
            documents: iterable of (id, data, tags)
            buffer: file path used for memmap buffer

        Returns:
            (document ids, dimensions, embeddings)
        """

        dtype = np.uint8 if self.qbits else np.float32

        return self.model.vectors(self.stream(documents), self.batch, self.checkpoint, buffer, dtype)

    def ids(self, documents):
        """
        Runs an ids transform operation when dense indexing is disabled.

        Args:
            documents: iterable of (id, data, tags)

        Returns:
            document ids
        """

        ids = []
        for uid, _, _ in self.stream(documents):
            ids.append(uid)

        self.config["offset"] = self.offset

        return ids

    def stream(self, documents):
        """
        This method does two things:

        1. Filter and yield data to vectorize
        2. Batch and load original documents into enabled data stores (database, graph, scoring)

        Documents are yielded for vectorization if one of the following is True:
            - dict with a text or object field
            - not a dict

        Otherwise, documents are only batched and inserted into data stores

        Args:
            documents: iterable collection (id, data, tags)
        """

        batch, offset = [], 0

        for document in documents:
            if isinstance(document[1], dict):
                if not self.indexing and not document[1].get(self.text):
                    document[1][self.text] = str(document[0])

                if self.text in document[1]:
                    yield (document[0], document[1][self.text], document[2])
                    offset += 1
                elif self.object in document[1]:
                    yield (document[0], document[1][self.object], document[2])
                    offset += 1
            else:
                yield document
                offset += 1

            batch.append(document)
            if len(batch) == self.batch:
                self.load(batch, offset)
                batch, offset = [], 0

        if batch:
            self.load(batch, offset)

    def load(self, batch, offset):
        """
        Loads a document batch. This method deletes existing ids from an embeddings index and
        loads into enabled data stores (database, graph, scoring).

        Args:
            batch: list of (id, data, tags)
            offset: index offset for batch
        """

        if self.action == Action.UPSERT:
            deletes = [uid for uid, _, _ in batch if uid not in self.deletes]
            if deletes:
                self.delete(deletes)

                self.deletes.update(deletes)

        if self.database and self.action != Action.REINDEX:
            self.database.insert(batch, self.offset)

        if self.scoring:
            self.scoring.insert(batch, self.offset, self.checkpoint)

        if self.indexes:
            self.indexes.insert(batch, self.offset, self.checkpoint)

        if self.graph:
            self.graph.insert(batch, self.offset)

        self.offset += offset



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import types

import numpy as np

from ...utilities import Resolver
from ..base import Vectors


class External(Vectors):
    """
    Builds vectors using an external method. This can be a local function or an external API call.
    """

    def __init__(self, config, scoring, models):
        super().__init__(config, scoring, models)

        self.transform = self.resolve(config.get("transform"))

    def loadmodel(self, path):
        return None

    def encode(self, data, category=None):
        if self.transform and data and not isinstance(data[0], np.ndarray):
            data = self.transform(data)

        return data.astype(np.float32) if isinstance(data, np.ndarray) else np.array(data, dtype=np.float32)

    def resolve(self, transform):
        """
        Resolves a transform function.

        Args:
            transform: transform function

        Returns:
            resolved transform function
        """

        if transform:
            transform = Resolver()(transform) if transform and isinstance(transform, str) else transform

            transform = transform if isinstance(transform, types.FunctionType) else transform()

        return transform



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Task


class StreamTask(Task):
    """
    Task that calls a task action and yields results.
    """

    def register(self, batch=False):
        """
        Adds stream parameters to task.

        Args:
            batch: all elements are passed to a single action call if True, otherwise
                        an action call is executed per element, defaults to False
        """

        # pylint: disable=W0201
        self.batch = batch

    def __call__(self, elements, executor=None):
        for action in self.action:
            if self.batch:
                yield from action(elements)
            else:
                for x in elements:
                    yield from action(x)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO


class Encoder:
    """
    Encodes and decodes object content. The base encoder works only with byte arrays. It can be extended
    to encode different datatypes.
    """

    def encode(self, obj):
        """
        Encodes an object to a byte array using the encoder.

        Args:
            obj: object to encode

        Returns:
            encoded object as a byte array
        """

        return obj

    def decode(self, data):
        """
        Decodes input byte array into an object using this encoder.

        Args:
            data: encoded data

        Returns:
            decoded object
        """

        return BytesIO(data) if data else None



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil


class Recovery:
    """
    Vector embeddings recovery. This class handles streaming embeddings from a vector checkpoint file.
    """

    def __init__(self, checkpoint, vectorsid, load):
        """
        Creates a Recovery instance.

        Args:
            checkpoint: checkpoint directory
            vectorsid: vectors uid for current configuration
            load: load embeddings method
        """

        self.spool, self.path, self.load = None, None, load

        path = f"{checkpoint}/{vectorsid}"
        if os.path.exists(path):
            self.path = f"{checkpoint}/recovery"

            shutil.copyfile(path, self.path)

            # pylint: disable=R1732
            self.spool = open(self.path, "rb")

    def __call__(self):
        """
        Reads and returns the next batch of embeddings.

        Returns
            batch of embeddings
        """

        try:
            return self.load(self.spool) if self.spool else None
        except EOFError:
            self.spool.close()
            os.remove(self.path)

            self.spool, self.path = None, None

            return None



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .errors import SerializeError
from .factory import SerializeFactory


class Serializer:
    """
    Methods to serialize and deserialize data.
    """

    @staticmethod
    def load(path):
        """
        Loads data from path. This method first tries to load the default serialization format.
        If that fails, it will fallback to pickle format for backwards-compatability purposes.

        Note that loading pickle files requires the env variable `ALLOW_PICKLE=True`.

        Args:
            path: data to load

        Returns:
            data
        """

        try:
            return SerializeFactory.create().load(path)
        except SerializeError:
            return SerializeFactory.create("pickle").load(path)

    @staticmethod
    def save(data, path):
        """
        Saves data to path.

        Args:
            data: data to save
            path: output path
        """

        SerializeFactory.create().save(data, path)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import re

try:
    from grandcypher import GrandCypher  # type:ignore

    GRANDCYPHER = True
except ImportError:
    GRANDCYPHER = False

logger = logging.getLogger(__name__)


class Query:
    """
    Runs openCypher graph queries using the GrandCypher library. This class also supports search functions.
    """

    SIMILAR = "__SIMILAR__"

    def __init__(self):
        """
        Create a new graph query instance.
        """

        if not GRANDCYPHER:
            raise ImportError('GrandCypher is not available - install "graph" extra to enable')

    def __call__(self, graph, query, limit):
        """
        Runs a graph query.

        Args:
            graph: graph instance
            query: graph query, can be a full query string or a parsed query dictionary
            limit: number of results

        Returns:
            results
        """

        attributes, uids = None, None

        if isinstance(query, dict):
            query, attributes, uids = self.build(query)

        if uids:
            graph = self.filter(graph, attributes, uids)

        logger.debug(query)

        return GrandCypher(graph.backend, limit if limit else 3).run(query)

    def isquery(self, queries):
        """
        Checks a list of queries to see if all queries are openCypher queries.

        Args:
            queries: list of queries to check

        Returns:
            True if all queries are openCypher queries
        """

        return all(query and query.strip().startswith("MATCH ") and "RETURN " in query for query in queries)

    def parse(self, query):
        """
        Parses a graph query. This method supports parsing search functions and replacing them with placeholders.

        Args:
            query: graph query

        Returns:
            parsed query as a dictionary
        """

        where, limit, nodes, similar = None, None, [], []

        match = re.search(r"where(.+?)return", query, flags=re.DOTALL | re.IGNORECASE)
        if match:
            where = match.group(1).strip()

        match = re.search(r"limit\s+(\d+)", query, flags=re.DOTALL | re.IGNORECASE)
        if match:
            limit = match.group(1)

        for x, match in enumerate(re.finditer(r"similar\((.+?)\)", query, flags=re.DOTALL | re.IGNORECASE)):
            query = query.replace(match.group(0), f"{Query.SIMILAR}{x}")

            params = [param.strip().replace("'", "").replace('"', "") for param in match.group(1).split(",")]
            nodes.append(params[0])
            similar.append(params[1:])

        return {
            "query": query,
            "where": where,
            "limit": limit,
            "nodes": nodes,
            "similar": similar,
        }

    def build(self, parse):
        """
        Constructs a full query from a parsed query. This method supports substituting placeholders with search results.

        Args:
            parse: parsed query

        Returns:
            graph query
        """

        query, attributes, uids = parse["query"], {}, {}

        if "results" in parse:
            for x, result in enumerate(parse["results"]):
                node = parse["nodes"][x]

                attribute = f"match_{x}"
                clause = f"{node}.{attribute} > 0"

                query = query.replace(f"{Query.SIMILAR}{x}", f"{clause}")

                for uid, score in result:
                    if uid not in uids:
                        uids[uid] = score

                attributes[attribute] = result

        return query, attributes, uids.items()

    def filter(self, graph, attributes, uids):
        """
        Filters the input graph by uids. This method also adds similar match attributes.

        Args:
            graph: graph instance
            attributes: results by attribute matched
            uids: single list with all matching ids

        Returns:
            filtered graph
        """

        graph = graph.filter(uids)

        for attribute, result in attributes.items():
            for uid, score in result:
                graph.addattribute(uid, attribute, score)

        return graph



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

try:
    from bs4 import BeautifulSoup, NavigableString

    SOUP = True
except ImportError:
    SOUP = False

from ..base import Pipeline


class HTMLToMarkdown(Pipeline):
    """
    HTML to Markdown pipeline.

    Markdown formatting is applied for headings, blockquotes, lists, code, tables and text. Visual formatting is also
    included (bold, italic etc).

    This pipeline searches for the best node that has relevant text, often found with an article, main or body tag.
    """

    def __init__(self, paragraphs=False, sections=False):
        """
        Create a new Extract instance.

        Args:
            paragraphs: True if paragraph parsing enabled, False otherwise
            sections: True if section parsing enabled, False otherwise
        """

        if not SOUP:
            raise ImportError('HTMLToMarkdown pipeline is not available - install "pipeline" extra to enable')

        self.paragraphs = paragraphs
        self.sections = sections

    def __call__(self, html):
        """
        Transforms input HTML into Markdown formatted text.

        Args:
            html: input html

        Returns:
            markdown formatted text
        """

        soup = BeautifulSoup(html, features="html.parser")

        for script in soup.find_all(["script", "style"]):
            script.decompose()

        article = next((x for x in ["article", "main"] if soup.find(x)), None)

        nodes = []
        for node in soup.find_all(article if article else "body"):
            if not article or node.find("p"):
                nodes.append(self.process(node, article))

        return "\n".join(self.metadata(soup) + nodes) if nodes else self.default(soup)

    def process(self, node, article):
        """
        Extracts text from a node. This method applies transforms for headings, blockquotes, lists, code, tables and text
        Page breaks are detected and reflected in the output text as a page break character.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            node text
        """

        if self.isheader(node):
            return self.header(node, article)

        if node.name in ("blockquote", "q"):
            return self.block(node)

        if node.name in ("ul", "ol"):
            return self.items(node, article)

        if node.name in ("code", "pre"):
            return self.code(node)

        if node.name == "table":
            return self.table(node, article)

        if node.name in ("aside",) + (() if article else ("header", "footer")):
            return ""

        page = node.name and node.get("class") and "page" in node.get("class")

        children = self.children(node)

        if self.iscontainer(node, children):
            texts = [self.process(node, article) for node in children]
            text = "\n".join(text for text in texts if text or not article)
        else:
            text = self.text(node, article)

        return f"{text}\f" if page and self.sections else text

    def metadata(self, node):
        """
        Builds a metadata section. The metadata section consists of the title and
        description fields.

        Args:
            node: input document node

        Returns:
            metadata as a list
        """

        title = node.find("title")
        metadata = [f"**{title.text.strip()}**"] if title and title.text else []

        description = node.find("meta", attrs={"name": "description"})
        if description and description["content"]:
            metadata.append(f"\n*{description['content'].strip()}*")

        if metadata:
            metadata.append("\f" if self.sections else "\n\n")

        return metadata

    def default(self, soup):
        """
        Default text handler when valid HTML isn't detected.

        Args:
            soup: BeautifulSoup object

        Returns:
            text
        """

        lines = []
        for line in soup.get_text().split("\n"):
            lines.append(f"\f{line}" if self.sections and re.search(r"^#+ ", line) else line)

        return "\n".join(lines)

    def text(self, node, article):
        """
        Text handler. This method flattens a node and it's children to text.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            node text
        """

        items = self.children(node)
        items = items if items else [node]

        texts = []
        for x in items:
            target, text = x if x.name else node, x.text

            if text.strip():
                if target.name in ("b", "strong"):
                    text = f"**{text.strip()}** "
                elif target.name in ("i", "em"):
                    text = f"*{text.strip()}* "
                elif target.name == "a":
                    text = f"[{text.strip()}]({target.get('href')}) "

            texts.append(text)

        text = "".join(texts)

        text = self.articletext(node, text) if article else text

        text = text if node.name and text else text.strip()

        return text

    def header(self, node, article):
        """
        Header handler. This method transforms a HTML heading into a Markdown formatted heading.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            heading as markdown
        """

        level = "#" * int(node.name[1])
        text = self.text(node, article)

        level = f"\f{level}" if self.sections else f"\n{level}"

        return f"{level} {text.lstrip()}" if text.strip() else ""

    def block(self, node):
        """
        Blockquote handler. This method transforms a HTML blockquote or q block into a Markdown formatted
        blockquote

        Args:
            node: input node

        Returns:
            block as markdown
        """

        text = "\n".join(f"> {x}" for x in node.text.strip().split("\n"))
        return f"{text}\n\n" if self.paragraphs else f"{text}\n"

    def items(self, node, article):
        """
        List handler. This method transforms a HTML ordered/unordered list into a Markdown formatted list.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            list as markdown
        """

        elements = []
        for x, element in enumerate(node.find_all("li")):
            prefix = "-" if node.name == "ul" else f"{x + 1}."

            text = self.process(element, article)

            if text:
                elements.append(f"{prefix} {text}")

        return "\n".join(elements)

    def code(self, node):
        """
        Code block handler. This method transforms a HTML pre or code block into a Markdown formatted
        code block.

        Args:
            node: input node

        Returns:
            code as markdown
        """

        text = f"```\n{node.text.strip()}\n```"
        return f"{text}\n\n" if self.paragraphs else f"{text}\n"

    def table(self, node, article):
        """
        Table handler. This method transforms a HTML table into a Markdown formatted table.

        Args:
            node: input node
            article: True if the main section node is an article

        Returns:
            table as markdown
        """

        elements, header = [], False

        rows = node.find_all("tr")
        for row in rows:
            columns = row.find_all(lambda tag: tag.name in ("th", "td"))

            elements.append(f"|{'|'.join(self.process(column, article) for column in columns)}|")

            if not header and len(rows) > 1:
                elements.append(f"{'|---' * len(columns)}|")
                header = True

        return "\n".join(elements)

    def iscontainer(self, node, children):
        """
        Analyzes a node and it's children to determine if this is a container element. A container
        element is defined as being a div, body, article or not having any string elements as children.

        Args:
            node: input node
            nodes: input node's children

        Returns:
            True if this is a container element, False otherwise
        """

        return children and (
            node.name in ("div", "body", "article") or not any(isinstance(x, NavigableString) for x in children)
        )

    def children(self, node):
        """
        Gets the node children, if available.

        Args:
            node: input node

        Returns:
            node children or None if not available
        """

        if node.name and node.contents:
            return [node for node in node.contents if node.name or node.text.strip()]

        return None

    def articletext(self, node, text):
        """
        Transforms node text using article parsing rules. Article parsing is designed to extract text content from web.
        It ignores navigation headers and other superfluous elements.

        Args:
            node: input node
            text: current text

        Returns:
            article text
        """

        valid = ("p", "th", "td", "li", "a", "b", "strong", "i", "em")

        valid = node.name in valid or (node.parent and node.parent.name in ("th", "td"))

        text = text if (valid or self.isheader(node)) and not self.islink(node) else ""
        if text:
            text = text.replace("\xa0\n", "\n\n")

            if node.name == "p":
                text = f"{text.strip()}\n\n" if self.paragraphs else f"{text.strip()}\n"

        return text

    def isheader(self, node):
        """
        Checks if node is a header node.

        Args:
            node: input node

        Returns:
            True if node is a header node, False otherwise
        """

        return node.name in ("h1", "h2", "h3", "h4", "h5", "h6")

    def islink(self, node):
        """
        Checks if node is a link node. This method does not consider links without tables as link nodes.

        Args:
            node: input node

        Returns:
            True if node is a link node, False otherwise
        """

        link, parent = False, node
        while parent:
            if parent.name == "a":
                link = True
                break

            parent = parent.parent

        return link and node.parent.name not in ("th", "td")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
import sqlite3
import sys
from array import array
from collections import Counter
from threading import RLock

import numpy as np


class Terms:
    """
    Builds, searches and stores memory efficient term frequency sparse arrays for a scoring instance.
    """

    CREATE_TERMS = """
        CREATE TABLE IF NOT EXISTS terms (
            term TEXT PRIMARY KEY,
            ids BLOB,
            freqs BLOB
        )
    """

    INSERT_TERM = "INSERT OR REPLACE INTO terms VALUES (?, ?, ?)"
    SELECT_TERMS = "SELECT ids, freqs FROM terms WHERE term = ?"

    CREATE_DOCUMENTS = """
        CREATE TABLE IF NOT EXISTS documents (
            indexid INTEGER PRIMARY KEY,
            id TEXT,
            deleted INTEGER,
            length INTEGER
        )
    """

    DELETE_DOCUMENTS = "DELETE FROM documents"
    INSERT_DOCUMENT = "INSERT OR REPLACE INTO documents VALUES (?, ?, ?, ?)"
    SELECT_DOCUMENTS = "SELECT indexid, id, deleted, length FROM documents ORDER BY indexid"

    def __init__(self, config, score, idf):
        """
        Creates a new terms index.

        Args:
            config: configuration
            score: score function
            idf: idf weights
        """

        self.config = config if isinstance(config, dict) else {}
        self.cachelimit = self.config.get("cachelimit", 250000000)
        self.cutoff = self.config.get("cutoff", 0.1)

        self.score, self.idf = score, idf

        self.ids, self.deletes, self.lengths = [], [], array("q")

        self.terms, self.cachesize = {}, 0

        self.connection, self.cursor, self.path = None, None, None

        self.lock = RLock()

    def insert(self, uid, terms):
        """
        Insert term into index.

        Args:
            uid: document id
            terms: document terms
        """

        self.initialize()

        indexid = len(self.ids)

        freqs, length = Counter(terms), len(terms)

        for term, count in freqs.items():
            self.add(indexid, term, count)

            self.cachesize += 16

        if self.cachesize >= self.cachelimit:
            self.index()

        self.ids.append(uid)
        self.lengths.append(length)

    def delete(self, ids):
        """
        Mark ids as deleted. This prevents deleted results from showing up in search results.
        The data is not removed from the underlying term frequency sparse arrays.

        Args:
            ids: ids to delete
        """

        self.deletes.extend([self.ids.index(i) for i in ids])

    def index(self):
        """
        Saves any remaining cached terms to the database.
        """

        for term, (nuids, nfreqs) in self.terms.items():
            uids, freqs = self.lookup(term)

            if uids:
                uids.extend(nuids)
                freqs.extend(nfreqs)
            else:
                uids, freqs = nuids, nfreqs

            if sys.byteorder == "big":
                uids.byteswap()
                freqs.byteswap()

            self.cursor.execute(Terms.INSERT_TERM, [term, uids.tobytes(), freqs.tobytes()])

        self.weights.cache_clear()

        self.terms, self.cachesize = {}, 0

    def search(self, terms, limit):
        """
        Searches term index a term-at-a-time. Each term frequency sparse array is retrieved
        and used to calculate term match scores.

        This method calculates term scores in two steps as shown below.

          1. Query and score less common term scores first
          2. Merge in common term scores for all documents matching the first query

        This is similar to the common terms query in Apache Lucene.

        Args:
            terms: query terms
            limit: maximum results

        Returns:
            list of (id, score)
        """

        scores = np.zeros(len(self.ids), dtype=np.float32)

        terms, skipped, hasscores = Counter(terms), {}, False
        for term, freq in terms.items():
            uids, weights = self.weights(term)
            if uids is not None:
                if len(uids) <= self.cutoff * len(self.ids):
                    scores[uids] += freq * weights

                    hasscores = True
                else:
                    skipped[term] = freq

        return self.topn(scores, limit, hasscores, skipped)

    def count(self):
        """
        Number of elements in the scoring index.

        Returns:
            count
        """

        return len(self.ids) - len(self.deletes)

    def load(self, path):
        """
        Loads terms database from path. This method loads document attributes into memory.

        Args:
            path: path to read terms database
        """

        self.connection = self.connect(path)
        self.cursor = self.connection.cursor()
        self.path = path

        self.ids, self.deletes, self.lengths = [], [], array("q")

        self.cursor.execute(Terms.SELECT_DOCUMENTS)
        for indexid, uid, deleted, length in self.cursor:
            self.ids.append(uid)

            if deleted:
                self.deletes.append(indexid)

            self.lengths.append(length)

        if all(uid.isdigit() for uid in self.ids):
            self.ids = [int(uid) for uid in self.ids]

        self.weights.cache_clear()

    def save(self, path):
        """
        Saves terms database to path. This method creates or replaces document attributes into the database.

        Args:
            path: path to write terms database
        """

        self.cursor.execute(Terms.DELETE_DOCUMENTS)

        for i, uid in enumerate(self.ids):
            self.cursor.execute(Terms.INSERT_DOCUMENT, [i, uid, 1 if i in self.deletes else 0, self.lengths[i]])

        if not self.path:
            self.connection.commit()

            connection = self.copy(path)

            self.connection.close()

            self.connection = connection
            self.cursor = self.connection.cursor()
            self.path = path

        elif self.path == path:
            self.connection.commit()

        else:
            self.copy(path).close()

    def close(self):
        """
        Close and free resources used by this instance.
        """

        if self.connection:
            self.connection.close()

    def initialize(self):
        """
        Creates connection and initial database schema if no connection exists.
        """

        if not self.connection:
            self.connection = self.connect()
            self.cursor = self.connection.cursor()

            self.cursor.execute(Terms.CREATE_TERMS)
            self.cursor.execute(Terms.CREATE_DOCUMENTS)

    def connect(self, path=""):
        """
        Creates a new term database connection.

        Args:
            path: path to term database file

        Returns:
            connection
        """

        connection = sqlite3.connect(path, check_same_thread=False)

        if self.config.get("wal"):
            connection.execute("PRAGMA journal_mode=WAL")

        return connection

    def copy(self, path):
        """
        Copies content from current terms database into target.

        Args:
            path: target database path

        Returns:
            new database connection
        """

        if os.path.exists(path):
            os.remove(path)

        connection = self.connect(path)

        if self.connection.in_transaction:
            for sql in self.connection.iterdump():
                connection.execute(sql)
        else:
            self.connection.backup(connection)

        return connection

    def add(self, indexid, term, freq):
        """
        Adds a term frequency entry.

        Args:
            indexid: internal index id
            term: term
            freq: term frequency
        """

        if term not in self.terms:
            self.terms[term] = (array("q"), array("q"))

        ids, freqs = self.terms[term]
        ids.append(indexid)
        freqs.append(freq)

    def lookup(self, term):
        """
        Retrieves a term frequency sparse array.

        Args:
            term: term to lookup

        Returns:
            term frequency sparse array
        """

        uids, freqs = None, None

        result = self.cursor.execute(Terms.SELECT_TERMS, [term]).fetchone()
        if result:
            uids, freqs = (array("q"), array("q"))
            uids.frombytes(result[0])
            freqs.frombytes(result[1])

            if sys.byteorder == "big":
                uids.byteswap()
                freqs.byteswap()

        return uids, freqs

    @functools.lru_cache(maxsize=500)  # noqa: B019
    def weights(self, term):
        """
        Computes a term weights sparse array for term. This method is wrapped with a least recently used cache,
        which will return common term weights from the cache.

        Args:
            term: term

        Returns:
            term weights sparse array
        """

        lengths = np.frombuffer(self.lengths, dtype=np.int64)

        with self.lock:
            uids, freqs = self.lookup(term)
            weights = None

        if uids:
            uids = np.frombuffer(uids, dtype=np.int64)
            weights = self.score(np.frombuffer(freqs, dtype=np.int64), self.idf[term], lengths[uids]).astype(np.float32)

        return uids, weights

    def topn(self, scores, limit, hasscores, skipped):
        """
        Get topn scores from an partial scores array.

        Args:
            scores: partial scores array with scores for less common terms
            limit: maximum results
            hasscores: True if partial scores array has any nonzero scores, False otherwise
            skipped: terms skipped in initial query

        Returns:
            topn scores
        """

        topn = min(len(scores), limit * 5)

        matches = self.candidates(scores, topn)

        self.merge(scores, matches, hasscores, skipped)

        if not hasscores:
            matches = self.candidates(scores, topn)

        matches = matches[np.argsort(-scores[matches])]

        return [(self.ids[x], float(scores[x])) for x in matches[:limit] if scores[x] > 0]

    def merge(self, scores, matches, hasscores, terms):
        """
        Merges common term scores into scores array.

        Args:
            scores: partial scores array
            matches: current matches, if any
            hasscores: True if scores has current matches, False otherwise
            terms: common terms
        """

        for term, freq in terms.items():
            uids, weights = self.weights(term)

            if hasscores:
                indices = np.searchsorted(uids, matches)

                indices = [x for i, x in enumerate(indices) if x < len(uids) and uids[x] == matches[i]]

                uids, weights = uids[indices], weights[indices]

            scores[uids] += freq * weights

    def candidates(self, scores, topn):
        """
        Gets the topn scored candidates. This method ignores deleted documents.

        Args:
            scores: scores array
            topn: topn elements

        Returns:
            topn scored candidates
        """

        scores[self.deletes] = 0

        return np.argpartition(scores, -topn)[-topn:]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from ...utilities import Resolver


class TaskFactory:
    """
    Task factory. Creates new Task instances.
    """

    @staticmethod
    def get(task):
        """
        Gets a new instance of task class.

        Args:
            task: Task instance class

        Returns:
            Task class
        """

        if "." not in task:
            task = ".".join(__name__.split(".")[:-1]) + "." + task.capitalize() + "Task"

        return Resolver()(task)

    @staticmethod
    def create(config, task):
        """
        Creates a new Task instance.

        Args:
            config: Task configuration
            task: Task instance class

        Returns:
            Task
        """

        if "args" in config:
            args = config.pop("args")
            action = config["action"]
            if action:
                if isinstance(action, list):
                    config["action"] = [Partial.create(a, args[i]) for i, a in enumerate(action)]
                else:
                    config["action"] = lambda x: action(x, **args) if isinstance(args, dict) else action(x, *args)

        return TaskFactory.get(task)(**config)


class Partial(functools.partial):
    """
    Modifies functools.partial to prepend arguments vs append.
    """

    @staticmethod
    def create(action, args):
        """
        Creates a new Partial function.

        Args:
            action: action to execute
            args: arguments

        Returns:
            Partial
        """

        return Partial(action, **args) if isinstance(args, dict) else Partial(action, *args) if args else Partial(action)

    def __call__(self, *args, **kwargs):
        kw = self.keywords.copy()
        kw.update(kwargs)

        return self.func(*(args + self.args), **kw)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Data


class Sequences(Data):
    """
    Tokenizes sequence-sequence datasets as input for training sequence-sequence models
    """

    def __init__(self, tokenizer, columns, maxlength, prefix):
        """
        Creates a new instance for tokenizing Sequences training data.

        Args:
            tokenizer: model tokenizer
            columns: tuple of columns to use for text/label
            maxlength: maximum sequence length
            prefix: source prefix
        """

        super().__init__(tokenizer, columns, maxlength)

        if not self.columns:
            self.columns = ("source", "target")

        self.prefix = prefix

    def process(self, data):
        source, target = self.columns

        source = [self.prefix + x if self.prefix else x for x in data[source]]
        inputs = self.tokenizer(source, max_length=self.maxlength, padding=False, truncation=True)

        with self.tokenizer.as_target_tokenizer():
            targets = self.tokenizer(data[target], max_length=self.maxlength, padding=False, truncation=True)

        inputs["labels"] = targets["input_ids"]

        return inputs



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

try:
    from libcloud.storage.providers import get_driver  # type:ignore

    LIBCLOUD = True
except ImportError:
    LIBCLOUD = False

from .base import Task


class StorageTask(Task):
    """
    Task that processes object storage buckets. Supports local and cloud providers in Apache libcloud.
    """

    PREFIX = r"(\w+):\/\/.*"
    PATH = r"\w+:\/\/(.*)"

    def register(self, key=None, secret=None, host=None, port=None, token=None, region=None):
        """
        Checks if required dependencies are installed. Reads in cloud storage parameters.

        Args:
            key: provider-specific access key
            secret: provider-specific access secret
            host: server host name
            port: server port
            token: temporary session token
            region: storage region
        """

        if not LIBCLOUD:
            raise ImportError('StorageTask is not available - install "workflow" extra to enable')

        # pylint: disable=W0201
        self.key = key
        self.secret = secret
        self.host = host
        self.port = port
        self.token = token
        self.region = region

    def __call__(self, elements, executor=None):
        outputs = []
        for element in elements:
            if self.matches(element):
                outputs.extend(super().__call__(self.list(element), executor))
            else:
                outputs.append(element)

        return outputs

    def matches(self, element):
        """
        Determines if this element is a storage element.

        Args:
            element: input storage element

        Returns:
            True if this is a storage element
        """

        return re.match(StorageTask.PREFIX, self.upack(element, True).lower())

    def list(self, element):
        """
        Gets a list of urls for a object container.

        Args:
            element: object container

        Returns:
            list of urls
        """

        provider = re.sub(StorageTask.PREFIX, r"\1", element.lower())
        path = re.sub(StorageTask.PATH, r"\1", element)

        key = self.key if self.key is not None else os.environ.get("ACCESS_KEY")
        secret = self.secret if self.secret is not None else os.environ.get("ACCESS_SECRET")

        key, container = (os.path.dirname(path), os.path.basename(path)) if key is None else (key, path)

        parts = container.split("/", 1)
        container, prefix = (parts[0], parts[1]) if len(parts) > 1 else (container, None)

        driver = get_driver(provider)

        client = driver(
            key,
            secret,
            **{field: getattr(self, field) for field in ["host", "port", "region", "token"] if getattr(self, field)},
        )

        container = client.get_container(container_name=container)
        return [
            client.get_object_cdn_url(obj) for obj in client.list_container_objects(container=container, prefix=prefix)
        ]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .token import Token


class Expression:
    """
    Parses expression statements and runs a set of substitution/formatting rules.
    """

    def __init__(self, resolver, tolist):
        """
        Creates a new expression parser.

        Args:
            resolver: function to call to resolve query column names with database column names
            tolist: outputs expression lists if True, text if False
        """

        self.resolver = resolver
        self.tolist = tolist

    def __call__(self, tokens, alias=False, aliases=None, similar=None):
        """
        Parses and formats a list of tokens as follows:
            - Replaces query column names with database column names
            - Adds similar query placeholders and extracts similar function parameters
            - Rewrites expression and returns

        Args:
            tokens: input expression
            alias: if True, column aliases should be generated and added to aliases dict
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
            similar: list of similar queries, if present new similar queries are appended to this list

        Returns:
            rewritten clause
        """

        transformed = self.process(list(tokens), alias, aliases, similar)

        if alias and not self.tolist:
            return self.buildalias(transformed, tokens, aliases)

        return self.buildlist(transformed) if self.tolist is True else self.buildtext(transformed)

    def process(self, tokens, alias, aliases, similar):
        """
        Replaces query column names with database column names, adds similar query placeholders and
        extracts similar function parameters.

        Args:
            tokens: input expression
            alias: if True, column aliases should be generated and added to aliases dict
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
            similar: list of similar queries, if present new similar queries are appended to this list

        Returns:
            transformed tokens
        """

        index, iterator = 0, ((x, token) for x, token in enumerate(tokens) if not Token.isdistinct(token))
        for x, token in iterator:
            if Token.isseparator(token):
                index += 1

            elif Token.isbracket(token):
                self.bracket(iterator, tokens, x)

            elif Token.issimilar(tokens, x, similar):
                self.similar(iterator, tokens, x, similar)

            elif Token.isfunction(tokens, x):
                self.function(iterator, tokens, token, aliases, similar)

            elif Token.isalias(tokens, x, alias):
                self.alias(iterator, tokens, x, aliases, index)

            elif Token.isattribute(tokens, x):
                self.attribute(tokens, x, aliases)

            elif Token.iscompound(tokens, x):
                self.compound(iterator, tokens, x, aliases, similar)

        return [token for token in tokens if token]

    def buildtext(self, tokens):
        """
        Builds a new expression from tokens. This method applies a set of rules to generate whitespace between tokens.

        Args:
            tokens: input expression

        Returns:
            expression text
        """

        text = ""
        for token in tokens:
            text += Token.wrapspace(text, token)

        return text.strip()

    def buildlist(self, tokens):
        """
        Builds a new expression from tokens. This method returns a list of expression components.
        These components can be joined together on commas to form a text expression.

        Args:
            tokens: input expression

        Returns:
            expression list
        """

        parts, current, parens, brackets = [], [], 0, 0

        for token in tokens:
            if token == "," and not parens and not brackets:
                parts.append(self.buildtext(current))
                current = []
            else:
                if token == "(":
                    parens += 1
                elif token == ")":
                    parens -= 1
                elif token == "[":
                    brackets += 1
                elif token == "]":
                    brackets -= 1
                elif Token.issortorder(token):
                    token = f" {token}"
                current.append(token)

        if current:
            parts.append(self.buildtext(current))

        return parts

    def buildalias(self, transformed, tokens, aliases):
        """
        Builds new alias text expression from transformed and input tokens.

        Args:
            transformed: transformed tokens
            tokens: original input tokens
            aliases: dict of column aliases

        Returns:
            alias text expression
        """

        transformed = self.buildlist(transformed)
        tokens = self.buildlist(tokens)

        expression = []
        for x, token in enumerate(transformed):
            if x not in aliases.values():
                alias = tokens[x]

                if not any(Token.isoperator(t) for t in alias) and alias[0] in ("[", "(") and alias[-1] in ("]", ")"):
                    alias = alias[1:-1]

                values = alias.split()
                if len(values) > 0 and Token.isdistinct(values[0]):
                    alias = " ".join(values[1:])

                token = self.resolver(token, alias)

            expression.append(token)

        return ", ".join(expression)

    def bracket(self, iterator, tokens, x):
        """
        Consumes a [bracket] expression.

        Args:
            iterator: tokens iterator
            tokens: input tokens
            x: current position
        """

        params = []

        token = tokens[x]
        tokens[x] = None

        brackets = 1

        while token and (token != "]" or brackets > 0):
            x, token = next(iterator, (None, None))

            if token == "[":
                brackets += 1
            elif token == "]":
                brackets -= 1

            if token != "]" or brackets > 0:
                params.append(token)

            tokens[x] = None

        tokens[x] = self.resolve(self.buildtext(params), None)

    def similar(self, iterator, tokens, x, similar):
        """
        Substitutes a similar() function call with a placeholder that can later be used to add
        embeddings query results as a filter.

        Args:
            iterator: tokens iterator
            tokens: input tokens
            x: current position
            similar: list where similar function call parameters are stored
        """

        params = []

        token = tokens[x]
        tokens[x] = None

        while token and token != ")":
            x, token = next(iterator, (None, None))
            if token and token not in ["(", ",", ")"]:
                params.append(token.replace("'", "").replace('"', ""))

            tokens[x] = None

        tokens[x] = f"{Token.SIMILAR_TOKEN}{len(similar)}"

        similar.append(params)

    def function(self, iterator, tokens, token, aliases, similar):
        """
        Resolves column names within the function's parameters.

        Args:
            iterator: tokens iterator
            tokens: input tokens
            token: current token
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
            similar: list where similar function call parameters are stored
        """

        while token and token != ")":
            x, token = next(iterator, (None, None))

            if Token.isbracket(token):
                self.bracket(iterator, tokens, x)

            elif Token.issimilar(tokens, x, similar):
                self.similar(iterator, tokens, x, similar)

            elif Token.isfunction(tokens, x):
                self.function(iterator, tokens, token, aliases, similar)

            elif Token.isattribute(tokens, x):
                self.attribute(tokens, x, aliases)

            elif Token.iscompound(tokens, x):
                self.compound(iterator, tokens, x, aliases, similar)

    def alias(self, iterator, tokens, x, aliases, index):
        """
        Reads an alias clause and stores it in aliases.

        Args:
            iterator: tokens iterator
            tokens: input tokens
            x: current position
            aliases: dict where aliases are stored - stores {alias: clause index}
            index: clause index, used to match aliases with columns
        """

        token = tokens[x]

        if token in Token.ALIAS:
            x, token = next(iterator, (None, None))

        while x + 1 < len(tokens) and not Token.isseparator(Token.get(tokens, x + 1)):
            x, token = next(iterator, (None, None))

        aliases[Token.normalize(token)] = index

    def attribute(self, tokens, x, aliases):
        """
        Resolves an attribute column name.

        Args:
            tokens: input tokens
            x: current token position
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
        """

        tokens[x] = self.resolve(tokens[x], aliases)

    def compound(self, iterator, tokens, x, aliases, similar):
        """
        Resolves column names in a compound expression (left side <operator(s)> right side).

        Args:
            iterator: tokens iterator
            tokens: input tokens
            x: current token position
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
            similar: list where similar function call parameters are stored
        """

        if Token.iscolumn(tokens[x - 1]):
            tokens[x - 1] = self.resolve(tokens[x - 1], aliases)

        token = tokens[x]
        while token and Token.isoperator(token):
            x, token = next(iterator, (None, None))

        if token and Token.iscolumn(token):
            if Token.isfunction(tokens, x):
                self.function(iterator, tokens, token, aliases, similar)
            else:
                tokens[x] = self.resolve(token, aliases)

    def resolve(self, token, aliases):
        """
        Resolves this token's value if it is not an alias or a bind parameter.

        Args:
            token: token to resolve
            aliases: dict of generated aliases, if present these tokens should NOT be resolved

        Returns:
            resolved token value
        """

        if (aliases and Token.normalize(token) in aliases) or (token.startswith(":")):
            return token

        return self.resolver(token)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import urlparse

from ..utilities import Resolver
from .client import Client
from .duckdb import DuckDB
from .sqlite import SQLite


class DatabaseFactory:
    """
    Methods to create document databases.
    """

    @staticmethod
    def create(config):
        """
        Create a Database.

        Args:
            config: database configuration parameters

        Returns:
            Database
        """

        database = None

        content = config.get("content")

        if content is True:
            content = "sqlite"

        if content == "duckdb":
            database = DuckDB(config)
        elif content == "sqlite":
            database = SQLite(config)
        elif content:
            url = urlparse(content)
            if content == "client" or url.scheme:
                database = Client(config)
            else:
                database = DatabaseFactory.resolve(content, config)

        config["content"] = content

        return database

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            Database
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve database backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Models imports
"""

from .models import Models
from .onnx import OnnxModel
from .pooling import ClsPooling, MeanPooling, Pooling, PoolingFactory
from .registry import Registry
from .tokendetection import TokenDetection

__all__ = ("ClsPooling", "MeanPooling", "Models", "OnnxModel", "Pooling", "PoolingFactory", "Registry", "TokenDetection")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Data


class Questions(Data):
    """
    Tokenizes question-answering datasets as input for training question-answering models.
    """

    def __init__(self, tokenizer, columns, maxlength, stride):
        """
        Creates a new instance for tokenizing Questions training data.

        Args:
            tokenizer: model tokenizer
            columns: tuple of columns to use for question/context/answer
            maxlength: maximum sequence length
            stride: chunk size for splitting data for QA tasks
        """

        super().__init__(tokenizer, columns, maxlength)

        if not self.columns:
            self.columns = ("question", "context", "answers")

        self.question, self.context, self.answer = self.columns
        self.stride = stride
        self.rpad = tokenizer.padding_side == "right"

    def process(self, data):
        tokenized = self.tokenize(data)

        samples = tokenized.pop("overflow_to_sample_mapping")
        offsets = tokenized.pop("offset_mapping")

        tokenized["start_positions"] = []
        tokenized["end_positions"] = []

        for x, offset in enumerate(offsets):
            inputids = tokenized["input_ids"][x]
            clstoken = inputids.index(self.tokenizer.cls_token_id)

            sequences = tokenized.sequence_ids(x)

            answers = self.answers(data, samples[x])

            if len(answers["answer_start"]) == 0:
                tokenized["start_positions"].append(clstoken)
                tokenized["end_positions"].append(clstoken)
            else:
                startchar = answers["answer_start"][0]
                endchar = startchar + len(answers["text"][0])

                start = 0
                while sequences[start] != (1 if self.rpad else 0):
                    start += 1

                end = len(inputids) - 1
                while sequences[end] != (1 if self.rpad else 0):
                    end -= 1

                while start < len(offset) and offset[start][0] <= startchar:
                    start += 1
                tokenized["start_positions"].append(start - 1)

                while offset[end][1] >= endchar:
                    end -= 1
                tokenized["end_positions"].append(end + 1)

        return tokenized

    def tokenize(self, data):
        """
        Tokenizes batch of data

        Args:
            data: input data batch

        Returns:
            tokenized data
        """

        data[self.question] = [x.lstrip() for x in data[self.question]]

        return self.tokenizer(
            data[self.question if self.rpad else self.context],
            data[self.context if self.rpad else self.question],
            truncation="only_second" if self.rpad else "only_first",
            max_length=self.maxlength,
            stride=self.stride,
            return_overflowing_tokens=True,
            return_offsets_mapping=True,
            padding=True,
        )

    def answers(self, data, index):
        """
        Gets and formats an answer.

        Args:
            data: input examples
            index: answer index to retrieve

        Returns:
            answers dict
        """

        answers = data[self.answer][index]
        context = data[self.context][index]

        if not isinstance(answers, dict):
            if not answers:
                answers = {"text": [], "answer_start": []}
            else:
                answers = {"text": [answers], "answer_start": [context.index(answers)]}

        return answers



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Ids class for resolving internal ids for lists of ids in embeddings.
"""


class Ids:
    """
    Resolves internal ids for lists of ids in embeddings.

    Args:
        embeddings (Embeddings): Embeddings instance to resolve ids from.
    """

    def __init__(self, embeddings):
        """
        Create a new ids action.

        Args:
            embeddings: embeddings instance
        """

        self.database = embeddings.database
        self.ids = embeddings.ids

    def __call__(self, ids):
        """
        Resolve internal ids.

        Args:
            ids: ids

        Returns:
            internal ids
        """

        results = self.database.ids(ids) if self.database else self.scan(ids)

        ids = {}
        for iid, uid in results:
            if uid not in ids:
                ids[uid] = []
            ids[uid].append(iid)

        return ids

    def scan(self, ids):
        """
        Scans embeddings ids array for matches when content is disabled.

        Args:
            ids: search ids

        Returns:
            internal ids
        """

        indices = []
        for uid in ids:
            indices.extend([(index, value) for index, value in enumerate(self.ids) if uid == value])

        return indices



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Terms class for reducing query statements to keyword terms.
"""


class Terms:
    """
    Reduces a query statement down to keyword terms, extracting query text from similar clauses or returning the original query.

    Args:
        embeddings (Embeddings): Embeddings instance to extract terms from.
    """

    def __init__(self, embeddings):
        """
        Create a new terms action.

        Args:
            embeddings: embeddings instance
        """

        self.database = embeddings.database

    def __call__(self, queries):
        """
        Extracts keyword terms from a list of queries.

        Args:
            queries: list of queries

        Returns:
            list of queries reduced down to keyword term strings
        """

        if self.database:
            terms = []
            for query in queries:
                parse = self.database.parse(query)

                terms.append(" ".join(" ".join(s) for s in parse["similar"]))

            return terms

        return queries



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from tempfile import TemporaryDirectory

from .tar_compression import Tar
from .zip_compression import Zip


class Archive:
    """
    Base class for archive instances.
    """

    def __init__(self, directory=None):
        """
        Creates a new archive instance.

        Args:
            directory: directory to use as working directory, defaults to a temporary directory
        """

        self.directory = directory

    def isarchive(self, path):
        """
        Checks if path is an archive file based on the extension.

        Args:
            path: path to check

        Returns:
            True if the path ends with an archive extension, False otherwise
        """

        return path and any(path.lower().endswith(extension) for extension in [".tar.bz2", ".tar.gz", ".tar.xz", ".zip"])

    def path(self):
        """
        Gets the current working directory for this archive instance.

        Returns:
            archive working directory
        """

        if not self.directory:
            # pylint: disable=R1732
            self.directory = TemporaryDirectory()

        return self.directory.name if isinstance(self.directory, TemporaryDirectory) else self.directory

    def load(self, path, compression=None):
        """
        Extracts file at path to archive working directory.

        Args:
            path: path to archive file
            compression: compression format, infers from path if not provided
        """

        compress = self.create(path, compression)
        compress.unpack(path, self.path())

    def save(self, path, compression=None):
        """
        Archives files in archive working directory to file at path.

        Args:
            path: path to archive file
            compression: compression format, infers from path if not provided
        """

        output = os.path.dirname(path)
        if output:
            os.makedirs(output, exist_ok=True)

        compress = self.create(path, compression)
        compress.pack(self.path(), path)

    def create(self, path, compression):
        """
        Method to construct a Compress instance.

        Args:
            path: file path
            compression: compression format, infers using file extension if not provided

        Returns:
            Compress
        """

        compression = compression if compression else path.lower().split(".")[-1]

        return Zip() if compression == "zip" else Tar()

    def __repr__(self):
        """
        Returns a detailed string representation of the Archive instance.
        """
        class_name = self.__class__.__name__
        dir_info = self.directory if self.directory else "<TemporaryDirectory>"
        archive_types = ["tar", "zip"]
        return f"{class_name}(directory={dir_info!r}, types={archive_types})"

    def __str__(self):
        """
        Returns a user-friendly string representation of the Archive instance.
        """
        class_name = self.__class__.__name__
        dir_info = self.directory if self.directory else "temporary directory"
        return f"{class_name}(directory={dir_info}) [supports: tar, zip]"



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from queue import Queue
from threading import Thread

try:
    import sounddevice as sd  # type:ignore

    from .signal import SCIPY, Signal

    AUDIOSTREAM = SCIPY
except (ImportError, OSError):
    AUDIOSTREAM = False

from ..base import Pipeline


class AudioStream(Pipeline):
    """
    Threaded pipeline that streams audio segments to an output audio device. This pipeline is designed
    to run on local machines given that it requires access to write to an output device.
    """

    COMPLETE = (1, None)

    def __init__(self, rate=None):
        """
        Creates an AudioStream pipeline.

        Args:
            rate: optional target sample rate, otherwise uses input target rate with each audio segment
        """

        if not AUDIOSTREAM:
            raise ImportError(
                'AudioStream pipeline is not available - install "pipeline" extra to enable. '
                "Also check that the portaudio system library is available."
            )

        self.rate = rate

        self.queue = Queue()
        self.thread = Thread(target=self.play)
        self.thread.start()

    def __call__(self, segment):
        """
        Queues audio segments for the audio player.

        Args:
            segment: (audio, sample rate)|list

        Returns:
            segment
        """

        segments = [segment] if isinstance(segment, tuple) else segment

        for x in segments:
            self.queue.put(x)

        return segments[0] if isinstance(segment, tuple) else segments

    def wait(self):
        """
        Waits for all input audio segments to be played.
        """

        self.thread.join()

    def play(self):
        """
        Reads audio segments from queue. This method runs in a separate non-blocking thread.
        """

        audio, rate = self.queue.get()
        while not isinstance(audio, int) or (audio, rate) != AudioStream.COMPLETE:
            audio, rate = (Signal.resample(audio, rate, self.rate), self.rate) if self.rate else (audio, rate)

            sd.play(audio, rate, blocking=True)

            audio, rate = self.queue.get()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from ..compression import ArchiveFactory


class Cloud:
    """
    Base class for cloud providers. Cloud providers sync content between local and remote storage.
    """

    def __init__(self, config):
        """
        Creates a new cloud connection.

        Args:
            config: cloud configuration
        """

        self.config = config

    def exists(self, path=None):
        """
        Checks if path exists in cloud. If path is None, this method checks if the container exists.

        Args:
            path: path to check

        Returns:
            True if path or container exists, False otherwise
        """

        return self.metadata(path) is not None

    def metadata(self, path=None):
        """
        Returns metadata for path from cloud. If path is None, this method returns metadata
        for container.

        Args:
            path: retrieve metadata for this path

        Returns:
            path or container metadata if available, otherwise returns None
        """

        raise NotImplementedError

    def load(self, path=None):
        """
        Retrieves content from cloud and stores locally. If path is empty, this method retrieves
        all content in the container.

        Args:
            path: path to retrieve

        Returns:
            local path which can be different than input path
        """

        raise NotImplementedError

    def save(self, path):
        """
        Sends local content stored in path to cloud.

        Args:
            path: local path to sync
        """

        raise NotImplementedError

    def isarchive(self, path):
        """
        Check if path is an archive file.

        Args:
            path: path to check

        Returns:
            True if path ends with an archive extension, false otherwise
        """

        return ArchiveFactory.create().isarchive(path)

    def listfiles(self, path):
        """
        Lists files in path. If path is a file, this method returns a single element list
        containing path.

        Args:
            path: path to list

        Returns:
            List of files
        """

        if os.path.isdir(path):
            return [os.path.join(path, f) for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]

        return [path]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Workflow
from .execute import Execute
from .factory import WorkflowFactory
from .task import (
    ConsoleTask,
    ExportTask,
    ExtractorTask,
    FileTask,
    Formatter,
    ImageTask,
    RagTask,
    RetrieveTask,
    ServiceTask,
    StorageTask,
    StreamTask,
    Task,
    TaskFactory,
    TemplateFormatter,
    TemplateTask,
    UrlTask,
    WorkflowTask,
)

__all__ = (
    "ConsoleTask",
    "Execute",
    "ExportTask",
    "ExtractorTask",
    "FileTask",
    "Formatter",
    "ImageTask",
    "RagTask",
    "RetrieveTask",
    "ServiceTask",
    "StorageTask",
    "StreamTask",
    "Task",
    "TaskFactory",
    "TemplateFormatter",
    "TemplateTask",
    "UrlTask",
    "Workflow",
    "WorkflowFactory",
    "WorkflowTask",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import traceback
from datetime import datetime

try:
    from croniter import croniter  # type:ignore

    CRONITER = True
except ImportError:
    CRONITER = False

from .execute import Execute

logger = logging.getLogger(__name__)


class Workflow:
    """
    Base class for all workflows.
    """

    def __init__(self, tasks, batch=100, workers=None, name=None, stream=None):
        """
        Creates a new workflow. Workflows are lists of tasks to execute.

        Args:
            tasks: list of workflow tasks
            batch: how many items to process at a time, defaults to 100
            workers: number of concurrent workers
            name: workflow name
            stream: workflow stream processor
        """

        self.tasks = tasks
        self.batch = batch
        self.workers = workers
        self.name = name
        self.stream = stream

        self.workers = max(len(task.action) for task in self.tasks) if not self.workers else self.workers

    def __call__(self, elements):
        """
        Executes a workflow for input elements. This method returns a generator that yields transformed
        data elements.

        Args:
            elements: iterable data elements

        Returns:
            generator that yields transformed data elements
        """

        with Execute(self.workers) as executor:
            self.initialize()

            elements = self.stream(elements) if self.stream else elements

            for batch in self.chunk(elements):
                yield from self.process(batch, executor)

            self.finalize()

    def schedule(self, cron, elements, iterations=None):
        """
        Schedules a workflow using a cron expression and elements.

        Args:
            cron: cron expression
            elements: iterable data elements passed to workflow each call
            iterations: number of times to run workflow, defaults to run indefinitely
        """

        if not CRONITER:
            raise ImportError('Workflow scheduling is not available - install "workflow" extra to enable')

        logger.info("'%s' scheduler started with schedule %s", self.name, cron)

        maxiterations = iterations
        while iterations is None or iterations > 0:
            schedule = croniter(cron, datetime.now().astimezone()).get_next(datetime)
            logger.info("'%s' next run scheduled for %s", self.name, schedule.isoformat())
            time.sleep(schedule.timestamp() - time.time())

            # pylint: disable=W0703
            try:
                for _ in self(elements):
                    pass
            except Exception:
                logger.error(traceback.format_exc())

            if iterations is not None:
                iterations -= 1

        logger.info("'%s' max iterations (%d) reached", self.name, maxiterations)

    def initialize(self):
        """
        Runs task initializer methods (if any) before processing data.
        """

        for task in self.tasks:
            if task.initialize:
                task.initialize()

    def chunk(self, elements):
        """
        Splits elements into batches. This method efficiently processes both fixed size inputs and
        dynamically generated inputs.

        Args:
            elements: iterable data elements

        Returns:
            evenly sized batches with the last batch having the remaining elements
        """

        if hasattr(elements, "__len__") and hasattr(elements, "__getitem__"):
            for x in range(0, len(elements), self.batch):
                yield elements[x : x + self.batch]

        else:
            batch = []
            for x in elements:
                batch.append(x)

                if len(batch) == self.batch:
                    yield batch
                    batch = []

            if batch:
                yield batch

    def process(self, elements, executor):
        """
        Processes a batch of data elements.

        Args:
            elements: iterable data elements
            executor: execute instance, enables concurrent task actions

        Returns:
            transformed data elements
        """

        for x, task in enumerate(self.tasks):
            logger.debug("Running Task #%d", x)
            elements = task(elements, executor)

        yield from elements

    def finalize(self):
        """
        Runs task finalizer methods (if any) after all data processed.
        """

        for task in self.tasks:
            if task.finalize:
                task.finalize()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch

from .base import Pipeline


class Tensors(Pipeline):
    """
    Pipeline backed by a tensor processing framework. Currently supports PyTorch.
    """

    def quantize(self, model):
        """
        Quantizes input model and returns. This only is supported for CPU devices.

        Args:
            model: torch model

        Returns:
            quantized torch model
        """

        # pylint: disable=E1101
        return torch.quantization.quantize_dynamic(model, {torch.nn.Linear}, dtype=torch.qint8)

    def tensor(self, data):
        """
        Creates a tensor array.

        Args:
            data: input data

        Returns:
            tensor
        """

        # pylint: disable=E1102
        return torch.tensor(data)

    def context(self):
        """
        Defines a context used to wrap processing with the tensor processing framework.

        Returns:
            processing context
        """

        # pylint: disable=E1101
        return torch.no_grad()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3

from .embedded import Embedded


class SQLite(Embedded):
    """
    Database instance backed by SQLite.
    """

    def connect(self, path=""):
        connection = sqlite3.connect(path, check_same_thread=False)

        if self.setting("wal"):
            connection.execute("PRAGMA journal_mode=WAL")

        return connection

    def getcursor(self):
        return self.connection.cursor()

    def rows(self):
        return self.cursor

    def addfunctions(self):
        if self.connection and self.functions:
            sqlite3.enable_callback_tracebacks(True)

            for name, argcount, fn in self.functions:
                self.connection.create_function(name, argcount, fn)

    def copy(self, path):
        if os.path.exists(path):
            os.remove(path)

        connection = self.connect(path)

        if self.connection.in_transaction:
            for sql in self.connection.iterdump():
                connection.execute(sql)
        else:
            self.connection.backup(connection)

        return connection



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Workflow
from .task import TaskFactory


class WorkflowFactory:
    """
    Workflow factory. Creates new Workflow instances.
    """

    @staticmethod
    def create(config, name):
        """
        Creates a new Workflow instance.

        Args:
            config: Workflow configuration
            name: Workflow name

        Returns:
            Workflow
        """

        tasks = []
        for tconfig in config["tasks"]:
            task = tconfig.pop("task") if "task" in tconfig else ""
            tasks.append(TaskFactory.create(tconfig, task))

        config["tasks"] = tasks

        if "stream" in config:
            sconfig = config["stream"]
            task = sconfig.pop("task") if "task" in sconfig else "stream"

            config["stream"] = TaskFactory.create(sconfig, task)

        return Workflow(**config, name=name)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from tempfile import TemporaryDirectory

try:
    import networkx as nx
    from networkx.algorithms.community import asyn_lpa_communities, greedy_modularity_communities, louvain_partitions
    from networkx.readwrite import json_graph

    NETWORKX = True
except ImportError:
    NETWORKX = False

from ..compression import ArchiveFactory
from ..serialization import SerializeError, SerializeFactory
from .base import Graph
from .query import Query


# pylint: disable=R0904
class NetworkX(Graph):
    """
    Graph instance backed by NetworkX.
    """

    def __init__(self, config):
        super().__init__(config)

        if not NETWORKX:
            raise ImportError('NetworkX is not available - install "graph" extra to enable')

    def create(self):
        return nx.Graph()

    def count(self):
        return self.backend.number_of_nodes()

    def scan(self, attribute=None, data=False):
        graph = self.backend

        if attribute:
            graph = nx.subgraph_view(self.backend, filter_node=lambda x: attribute in self.node(x))

        return graph.nodes(data=True) if data else graph

    def node(self, node):
        return self.backend.nodes.get(node)

    def addnode(self, node, **attrs):
        self.backend.add_node(node, **attrs)

    def addnodes(self, nodes):
        self.backend.add_nodes_from(nodes)

    def removenode(self, node):
        if self.hasnode(node):
            self.backend.remove_node(node)

    def hasnode(self, node):
        return self.backend.has_node(node)

    def attribute(self, node, field):
        return self.node(node).get(field) if self.hasnode(node) else None

    def addattribute(self, node, field, value):
        if self.hasnode(node):
            self.node(node)[field] = value

    def removeattribute(self, node, field):
        return self.node(node).pop(field, None) if self.hasnode(node) else None

    def edgecount(self):
        return self.backend.number_of_edges()

    def edges(self, node):
        edges = self.backend.adj.get(node)
        if edges:
            return dict(sorted(edges.items(), key=lambda x: x[1].get("weight", 0), reverse=True))

        return None

    def addedge(self, source, target, **attrs):
        self.backend.add_edge(source, target, **attrs)

    def addedges(self, edges):
        self.backend.add_edges_from(edges)

    def hasedge(self, source, target=None):
        if target is None:
            edges = self.backend.adj.get(source)
            return len(edges) > 0 if edges else False

        return self.backend.has_edge(source, target)

    def centrality(self):
        rank = nx.degree_centrality(self.backend)
        return dict(sorted(rank.items(), key=lambda x: x[1], reverse=True))

    def pagerank(self):
        rank = nx.pagerank(self.backend, weight="weight")
        return dict(sorted(rank.items(), key=lambda x: x[1], reverse=True))

    def showpath(self, source, target):
        # pylint: disable=E1121
        return nx.shortest_path(self.backend, source, target, self.distance)

    def isquery(self, queries):
        return Query().isquery(queries)

    def parse(self, query):
        return Query().parse(query)

    def search(self, query, limit=None, graph=False):
        results = Query()(self, query, limit)

        if graph:
            nodes = set()
            for column in results.values():
                for value in column:
                    if isinstance(value, list):
                        nodes.update([node for node in value if node and not isinstance(node, dict)])
                    elif isinstance(value, dict):
                        nodes.update(uid for uid, attr in self.scan(data=True) if attr["id"] == value["id"])
                    elif value is not None:
                        nodes.add(value)

            return self.filter(list(nodes))

        keys = list(results.keys())
        rows, count = [], len(results[keys[0]])

        for x in range(count):
            rows.append({str(key): results[key][x] for key in keys})

        return rows

    def communities(self, config):
        algorithm = config.get("algorithm")

        if algorithm == "greedy":
            communities = greedy_modularity_communities(
                self.backend, weight="weight", resolution=config.get("resolution", 100)
            )
        elif algorithm == "lpa":
            communities = asyn_lpa_communities(self.backend, weight="weight", seed=0)
        else:
            communities = self.louvain(config)

        return communities

    def load(self, path):
        try:
            data = SerializeFactory.create().load(path)

            self.backend = self.create()
            self.backend.add_nodes_from(data["nodes"])
            self.backend.add_edges_from(data["edges"])

            self.categories = data.get("categories")

            self.topics = data.get("topics")

        except SerializeError:
            self.loadtar(path)

    def save(self, path):
        SerializeFactory.create().save(
            {
                "nodes": [(uid, self.node(uid)) for uid in self.scan()],
                "edges": list(self.backend.edges(data=True)),
                "categories": self.categories,
                "topics": self.topics,
            },
            path,
        )

    def loaddict(self, data):
        self.backend = json_graph.node_link_graph(data, name="indexid")
        self.categories, self.topics = data.get("categories"), data.get("topics")

    def savedict(self):
        data = json_graph.node_link_data(self.backend, name="indexid")
        data["categories"] = self.categories
        data["topics"] = self.topics

        return data

    def louvain(self, config):
        """
        Runs the Louvain community detection algorithm.

        Args:
            config: topic configuration

        Returns:
            list of [ids] per community
        """

        level = config.get("level", "best")

        results = list(
            louvain_partitions(self.backend, weight="weight", resolution=config.get("resolution", 100), seed=0)
        )

        return results[0] if level == "first" else results[-1]

    # pylint: disable=W0613
    def distance(self, source, target, attrs):
        """
        Computes distance between source and target nodes using weight.

        Args:
            source: source node
            target: target node
            attrs: edge attributes

        Returns:
            distance between source and target
        """

        distance = max(1.0 - attrs["weight"], 0.0)
        return distance if distance >= 0.15 else 1.00

    def loadtar(self, path):
        """
        Loads a graph from the legacy TAR file.

        Args:
            path: path to graph
        """

        serializer = SerializeFactory.create("pickle")

        with TemporaryDirectory() as directory:
            archive = ArchiveFactory.create(directory)
            archive.load(path, "tar")

            self.backend = serializer.load(f"{directory}/graph")

            path = f"{directory}/categories"
            if os.path.exists(path):
                self.categories = serializer.load(path)

            path = f"{directory}/topics"
            if os.path.exists(path):
                self.topics = serializer.load(path)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile
from urllib.parse import urlparse
from urllib.request import urlretrieve

from .url import UrlTask


class RetrieveTask(UrlTask):
    """
    Task that retrieves urls (local or remote) to a local directory.
    """

    def register(self, directory=None, flatten=True):
        """
        Adds retrieve parameters to task.

        Args:
            directory: local directory used to store retrieved files
            flatten: flatten input directory structure, defaults to True
        """

        # pylint: disable=W0201
        if not directory:
            # pylint: disable=R1732
            self.tempdir = tempfile.TemporaryDirectory()
            directory = self.tempdir.name

        os.makedirs(directory, exist_ok=True)

        self.directory = directory
        self.flatten = flatten

    def prepare(self, element):
        path = urlparse(element).path

        if self.flatten:
            path = os.path.join(self.directory, os.path.basename(path))
        else:
            path = os.path.join(self.directory, os.path.normpath(path.lstrip("/")))
            directory = os.path.dirname(path)

            os.makedirs(directory, exist_ok=True)

        urlretrieve(element, path)

        return path



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from scipy.sparse import csr_matrix, vstack
    from sklearn.preprocessing import normalize
    from sklearn.utils.extmath import safe_sparse_dot

    SPARSE = True
except ImportError:
    SPARSE = False

from ...utilities import SparseArray
from ..base import Vectors


# pylint: disable=W0223
class SparseVectors(Vectors):
    """
    Base class for sparse vector models. Vector models transform input content into sparse arrays.
    """

    def __init__(self, config, scoring, models):
        if not SPARSE:
            raise ImportError('SparseVectors is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def encode(self, data, category=None):
        embeddings = super().encode(data, category)

        embeddings = embeddings.cpu().coalesce()
        indices = embeddings.indices().numpy()
        values = embeddings.values().numpy()

        return csr_matrix((values, indices), shape=embeddings.size())

    def vectors(self, documents, batchsize=500, checkpoint=None, buffer=None, dtype=None):
        ids, dimensions, batches, stream = self.index(documents, batchsize, checkpoint)

        embeddings = None
        with open(stream, "rb") as queue:
            for _ in range(batches):
                data = self.loadembeddings(queue)
                embeddings = vstack((embeddings, data)) if embeddings is not None else data

        return (ids, dimensions, embeddings)

    def dot(self, queries, data):
        return safe_sparse_dot(queries, data.T, dense_output=True).tolist()

    def loadembeddings(self, f):
        return SparseArray().load(f)

    def saveembeddings(self, f, embeddings):
        SparseArray().save(f, embeddings)

    def truncate(self, embeddings):
        raise ValueError("Truncate is not supported for sparse vectors")

    def normalize(self, embeddings):
        return normalize(embeddings, copy=False)

    def quantize(self, embeddings):
        raise ValueError("Quantize is not supported for sparse vectors")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utilities import Resolver
from .annoy import Annoy
from .faiss import Faiss
from .hnsw import HNSW
from .numpy import NumPy
from .pgvector import PGVector
from .sqlite import SQLite
from .torch import Torch


class ANNFactory:
    """
    Methods to create VectoreIndex indexes.
    """

    @staticmethod
    def create(config):
        """
        Create an VectoreIndex.

        Args:
            config: index configuration parameters

        Returns:
            VectoreIndex
        """

        ann = None
        backend = config.get("backend", "faiss")

        if backend == "annoy":
            ann = Annoy(config)
        elif backend == "faiss":
            ann = Faiss(config)
        elif backend == "hnsw":
            ann = HNSW(config)
        elif backend == "numpy":
            ann = NumPy(config)
        elif backend == "pgvector":
            ann = PGVector(config)
        elif backend == "sqlite":
            ann = SQLite(config)
        elif backend == "torch":
            ann = Torch(config)
        else:
            ann = ANNFactory.resolve(backend, config)

        config["backend"] = backend

        return ann

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            VectoreIndex
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve ann backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Translation module
"""

try:
    from staticvectors import StaticVectors

    STATICVECTORS = True
except ImportError:
    STATICVECTORS = False

from huggingface_hub.hf_api import HfApi
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer

from ...modeling import Models
from ..hfmodel import HFModel


class Translation(HFModel):
    """
    Translates text from source language into target language.
    """

    DEFAULT_LANG_DETECT = "neuml/language-id-quantized"

    def __init__(self, path=None, quantize=False, gpu=True, batch=64, langdetect=None, findmodels=True):
        """
        Constructs a new language translation pipeline.

        Args:
            path: optional path to model, accepts Hugging Face model hub id or local path,
                  uses default model for task if not provided
            quantize: if model should be quantized, defaults to False
            gpu: True/False if GPU should be enabled, also supports a GPU device id
            batch: batch size used to incrementally process content
            langdetect: set a custom language detection function, method must take a list of strings and return
                        language codes for each, uses default language detector if not provided
            findmodels: True/False if the Hugging Face Hub will be searched for source-target translation models
        """

        super().__init__(path if path else "facebook/m2m100_418M", quantize, gpu, batch)

        self.detector = None
        self.langdetect = langdetect
        self.findmodels = findmodels

        self.models = {}
        self.ids = None

    def __call__(self, texts, target="en", source=None, showmodels=False):
        """
        Translates text from source language into target language.

        This method supports texts as a string or a list. If the input is a string,
        the return type is string. If text is a list, the return type is a list.

        Args:
            texts: text|list
            target: target language code, defaults to "en"
            source: source language code, detects language if not provided

        Returns:
            list of translated text
        """

        values = [texts] if not isinstance(texts, list) else texts

        languages = self.detect(values) if not source else [source] * len(values)
        unique = set(languages)

        langdict = {}
        for x, lang in enumerate(languages):
            if lang not in langdict:
                langdict[lang] = []
            langdict[lang].append((x, values[x]))

        results = {}
        for language in unique:
            inputs = langdict[language]

            outputs = []
            for chunk in self.batch([text for _, text in inputs], self.batchsize):
                outputs.extend(self.translate(chunk, language, target, showmodels))

            for y, (x, _) in enumerate(inputs):
                if showmodels:
                    model, op = outputs[y]
                    results[x] = (op.strip(), language, model)
                else:
                    results[x] = outputs[y].strip()

        results = [results[x] for x in sorted(results)]
        return results[0] if isinstance(texts, str) else results

    def modelids(self):
        """
        Runs a query to get a list of available language models from the Hugging Face API.

        Returns:
            list of source-target language model ids
        """

        ids = [x.id for x in HfApi().list_models(author="Helsinki-NLP")] if self.findmodels else []
        return set(ids)

    def detect(self, texts):
        """
        Detects the language for each element in texts.

        Args:
            texts: list of text

        Returns:
            list of languages
        """

        if not self.langdetect or isinstance(self.langdetect, str):
            return self.defaultdetect(texts)

        return self.langdetect(texts)

    def defaultdetect(self, texts):
        """
        Default language detection model.

        Args:
            texts: list of text

        Returns:
            list of languages
        """

        if not self.detector:
            if not STATICVECTORS:
                raise ImportError('Language detection is not available - install "pipeline" extra to enable')

            path = self.langdetect if self.langdetect else Translation.DEFAULT_LANG_DETECT

            self.detector = StaticVectors(path)

        texts = [x.lower().replace("\n", " ").replace("\r\n", " ") for x in texts]

        return [x[0][0] for x in self.detector.predict(texts)]

    def translate(self, texts, source, target, showmodels=False):
        """
        Translates text from source to target language.

        Args:
            texts: list of text
            source: source language code
            target: target language code

        Returns:
            list of translated text
        """

        if source == target:
            return texts

        path, model, tokenizer = self.lookup(source, target)

        model.to(self.device)
        indices = None
        maxlength = Models.maxlength(model, tokenizer)

        with self.context():
            if hasattr(tokenizer, "lang_code_to_id"):
                source = self.langid(tokenizer.lang_code_to_id, source)
                target = self.langid(tokenizer.lang_code_to_id, target)

                tokenizer.src_lang = source
                tokens, indices = self.tokenize(tokenizer, texts)

                translated = model.generate(
                    **tokens, forced_bos_token_id=tokenizer.lang_code_to_id[target], max_length=maxlength
                )
            else:
                tokens, indices = self.tokenize(tokenizer, texts)
                translated = model.generate(**tokens, max_length=maxlength)

        translated = tokenizer.batch_decode(translated, skip_special_tokens=True)

        results, last = [], -1
        for x, i in enumerate(indices):
            v = (path, translated[x]) if showmodels else translated[x]
            if i == last:
                results[-1] += v
            else:
                results.append(v)

            last = i

        return results

    def lookup(self, source, target):
        """
        Retrieves a translation model for source->target language. This method caches each model loaded.

        Args:
            source: source language code
            target: target language code

        Returns:
            (model, tokenizer)
        """

        path = self.modelpath(source, target)
        if path not in self.models:
            self.models[path] = self.load(path)

        return (path,) + self.models[path]

    def modelpath(self, source, target):
        """
        Derives a translation model path given source and target languages.

        Args:
            source: source language code
            target: target language code

        Returns:
            model path
        """

        if self.ids is None:
            self.ids = self.modelids()

        template = "Helsinki-NLP/opus-mt-%s-%s"
        path = template % (source, target)
        if path in self.ids:
            return path

        if self.findmodels and target == "en":
            return template % ("mul", target)

        return self.path

    def load(self, path):
        """
        Loads a model specified by path.

        Args:
            path: model path

        Returns:
            (model, tokenizer)
        """

        model = AutoModelForSeq2SeqLM.from_pretrained(path)
        tokenizer = AutoTokenizer.from_pretrained(path)

        model = self.prepare(model)

        return (model, tokenizer)

    def langid(self, languages, target):
        """
        Searches a list of languages for a prefix match on target.

        Args:
            languages: list of languages
            target: target language code

        Returns:
            best match or None if no match found
        """

        for lang in languages:
            if lang.startswith(target):
                return lang

        return None



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Serialize:
    """
    Base class for Serialize instances. This class serializes data to files, streams and bytes.
    """

    def load(self, path):
        """
        Loads data from path.

        Args:
            path: input path

        Returns:
            deserialized data
        """

        with open(path, "rb") as handle:
            return self.loadstream(handle)

    def save(self, data, path):
        """
        Saves data to path.

        Args:
            data: data to save
            path: output path
        """

        with open(path, "wb") as handle:
            self.savestream(data, handle)

    def loadstream(self, stream):
        """
        Loads data from stream.

        Args:
            stream: input stream

        Returns:
            deserialized data
        """

        raise NotImplementedError

    def savestream(self, data, stream):
        """
        Saves data to stream.

        Args:
            data: data to save
            stream: output stream
        """

        raise NotImplementedError

    def loadbytes(self, data):
        """
        Loads data from bytes.

        Args:
            data: input bytes

        Returns:
            deserialized data
        """

        raise NotImplementedError

    def savebytes(self, data):
        """
        Saves data as bytes.

        Args:
            data: data to save

        Returns:
            serialized data
        """

        raise NotImplementedError



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from ..hfpipeline import HFPipeline


class Caption(HFPipeline):
    """
    Constructs captions for images.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        if not PIL:
            raise ImportError('Captions pipeline is not available - install "pipeline" extra to enable')

        super().__init__("image-to-text", path, quantize, gpu, model, **kwargs)

    def __call__(self, images):
        """
        Builds captions for images.

        This method supports a single image or a list of images. If the input is an image, the return
        type is a string. If text is a list, a list of strings is returned

        Args:
            images: image|list

        Returns:
            list of captions
        """

        values = [images] if not isinstance(images, list) else images

        values = [Image.open(image) if isinstance(image, str) else image for image in values]

        captions = []
        for result in self.pipeline(values):
            text = " ".join([r["generated_text"] for r in result]).strip()
            captions.append(text)

        return captions[0] if not isinstance(images, list) else captions



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from string import Formatter


class TemplateFormatter(Formatter):
    """
    Custom Formatter that requires each argument to be consumed.
    """

    def check_unused_args(self, used_args, args, kwargs):
        difference = set(kwargs).difference(used_args)
        if difference:
            raise KeyError(difference)
